import io
import typing as t
import tarfile
from pathlib import Path
from tempfile import NamedTemporaryFile
from functools import wraps
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor

import fs
import requests
from rich.live import Live
from simple_di import inject
from simple_di import Provide
from rich.panel import Panel
from rich.console import Group
from rich.console import ConsoleRenderable
from rich.progress import TaskID
from rich.progress import Progress
from rich.progress import BarColumn
from rich.progress import TextColumn
from rich.progress import SpinnerColumn
from rich.progress import DownloadColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn
from rich.progress import TransferSpeedColumn

from ..tag import Tag
from ..bento import Bento
from ..bento import BentoStore
from ..utils import calc_dir_size
from ..models import Model
from ..models import copy_model
from ..models import ModelStore
from ...exceptions import NotFound
from ...exceptions import BentoMLException
from ..configuration.containers import BentoMLContainer
from ..yatai_rest_api_client.config import get_current_yatai_rest_api_client
from ..yatai_rest_api_client.schemas import BentoApiSchema
from ..yatai_rest_api_client.schemas import LabelItemSchema
from ..yatai_rest_api_client.schemas import BentoRunnerSchema
from ..yatai_rest_api_client.schemas import BentoUploadStatus
from ..yatai_rest_api_client.schemas import CreateBentoSchema
from ..yatai_rest_api_client.schemas import CreateModelSchema
from ..yatai_rest_api_client.schemas import ModelUploadStatus
from ..yatai_rest_api_client.schemas import UpdateBentoSchema
from ..yatai_rest_api_client.schemas import BentoManifestSchema
from ..yatai_rest_api_client.schemas import ModelManifestSchema
from ..yatai_rest_api_client.schemas import FinishUploadBentoSchema
from ..yatai_rest_api_client.schemas import FinishUploadModelSchema
from ..yatai_rest_api_client.schemas import CreateBentoRepositorySchema
from ..yatai_rest_api_client.schemas import CreateModelRepositorySchema


class ObjectWrapper(object):
    def __getattr__(self, name: str) -> t.Any:
        return getattr(self._wrapped, name)

    def __setattr__(self, name: str, value: t.Any) -> None:
        return setattr(self._wrapped, name, value)

    def wrapper_getattr(self, name: str):
        """Actual `self.getattr` rather than self._wrapped.getattr"""
        return getattr(self, name)

    def wrapper_setattr(self, name: str, value: t.Any) -> None:
        """Actual `self.setattr` rather than self._wrapped.setattr"""
        return object.__setattr__(self, name, value)

    def __init__(self, wrapped: t.Any):
        """
        Thin wrapper around a given object
        """
        self.wrapper_setattr("_wrapped", wrapped)


class _CallbackIOWrapper(ObjectWrapper):
    def __init__(
        self,
        callback: t.Callable[[int], None],
        stream: t.BinaryIO,
        method: "t.Literal['read', 'write']" = "read",
    ):
        """
        Wrap a given `file`-like object's `read()` or `write()` to report
        lengths to the given `callback`
        """
        super().__init__(stream)
        func = getattr(stream, method)
        if method == "write":

            @wraps(func)
            def write(data: t.Union[bytes, bytearray], *args: t.Any, **kwargs: t.Any):
                res = func(data, *args, **kwargs)
                callback(len(data))
                return res

            self.wrapper_setattr("write", write)
        elif method == "read":

            @wraps(func)
            def read(*args: t.Any, **kwargs: t.Any):
                data = func(*args, **kwargs)
                callback(len(data))
                return data

            self.wrapper_setattr("read", read)
        else:
            raise KeyError("Can only wrap read/write methods")


# Just make type checker happy
class BinaryIOCast(io.BytesIO):
    def __init__(  # pylint: disable=useless-super-delegation
        self, *args: t.Any, **kwargs: t.Any
    ) -> None:
        super().__init__(*args, **kwargs)


CallbackIOWrapper: t.Type[BinaryIOCast] = t.cast(
    t.Type[BinaryIOCast], _CallbackIOWrapper
)


# Just make type checker happy
class ProgressCast(Progress):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

    def __rich__(self) -> t.Union[ConsoleRenderable, str]:  # pragma: no cover
        ...


ProgressWrapper: t.Type[ProgressCast] = t.cast(t.Type[ProgressCast], ObjectWrapper)


class YataiClient:
    log_progress = ProgressWrapper(
        Progress(
            TextColumn("{task.description}"),
        )
    )

    spinner_progress = ProgressWrapper(
        Progress(
            TextColumn("  "),
            TimeElapsedColumn(),
            TextColumn("[bold purple]{task.fields[action]}"),
            SpinnerColumn("simpleDots"),
        )
    )

    transmission_progress = ProgressWrapper(
        Progress(
            TextColumn("[bold blue]{task.description}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
        )
    )

    progress_group = Group(
        Panel(Group(log_progress, spinner_progress)), transmission_progress
    )

    @contextmanager
    def spin(self, *, text: str):
        task_id = self.spinner_progress.add_task("", action=text)
        try:
            yield
        finally:
            self.spinner_progress.stop_task(task_id)
            self.spinner_progress.update(task_id, visible=False)

    @inject
    def push_bento(
        self,
        bento: "Bento",
        *,
        force: bool = False,
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        with Live(self.progress_group):
            upload_task_id = self.transmission_progress.add_task(
                f'Pushing Bento "{bento.tag}"', start=False, visible=False
            )
            self._do_push_bento(
                bento, upload_task_id, force=force, model_store=model_store
            )

    @inject
    def _do_push_bento(
        self,
        bento: "Bento",
        upload_task_id: TaskID,
        *,
        force: bool = False,
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        yatai_rest_client = get_current_yatai_rest_api_client()
        name = bento.tag.name
        version = bento.tag.version
        if version is None:
            raise BentoMLException(f"Bento {bento.tag} version cannot be None")
        info = bento.info
        model_tags = [m.tag for m in info.models]
        with ThreadPoolExecutor(max_workers=max(len(model_tags), 1)) as executor:

            def push_model(model: "Model"):
                model_upload_task_id = self.transmission_progress.add_task(
                    f'Pushing model "{model.tag}"', start=False, visible=False
                )
                self._do_push_model(model, model_upload_task_id, force=force)

            futures = executor.map(
                push_model, (model_store.get(name) for name in model_tags)
            )
            list(futures)
        with self.spin(text=f'Fetching Bento repository "{name}"'):
            bento_repository = yatai_rest_client.get_bento_repository(
                bento_repository_name=name
            )
        if not bento_repository:
            with self.spin(text=f'Bento repository "{name}" not found, creating now..'):
                bento_repository = yatai_rest_client.create_bento_repository(
                    req=CreateBentoRepositorySchema(name=name, description="")
                )
        with self.spin(text=f'Try fetching Bento "{bento.tag}" from Yatai..'):
            remote_bento = yatai_rest_client.get_bento(
                bento_repository_name=name, version=version
            )
        if (
            not force
            and remote_bento
            and remote_bento.upload_status == BentoUploadStatus.SUCCESS
        ):
            self.log_progress.add_task(
                f'[bold blue]Push failed: Bento "{bento.tag}" already exists in Yatai'
            )
            return
        labels: t.List[LabelItemSchema] = [
            LabelItemSchema(key=key, value=value) for key, value in info.labels.items()
        ]
        apis: t.Dict[str, BentoApiSchema] = {}
        models = [str(m.tag) for m in info.models]
        runners = [
            BentoRunnerSchema(name=r.name, runner_type=r.runnable_type)
            for r in info.runners
        ]
        manifest = BentoManifestSchema(
            service=info.service,
            bentoml_version=info.bentoml_version,
            apis=apis,
            models=models,
            runners=runners,
            size_bytes=calc_dir_size(bento.path),
        )
        if not remote_bento:
            with self.spin(text=f'Registering Bento "{bento.tag}" with Yatai..'):
                yatai_rest_client.create_bento(
                    bento_repository_name=bento_repository.name,
                    req=CreateBentoSchema(
                        description="",
                        version=version,
                        build_at=info.creation_time,
                        manifest=manifest,
                        labels=labels,
                    ),
                )
        else:
            with self.spin(text=f'Updating Bento "{bento.tag}"..'):
                yatai_rest_client.update_bento(
                    bento_repository_name=bento_repository.name,
                    version=version,
                    req=UpdateBentoSchema(
                        manifest=manifest,
                        labels=labels,
                    ),
                )
        with self.spin(text=f'Getting a presigned upload url for "{bento.tag}" ..'):
            remote_bento = yatai_rest_client.presign_bento_upload_url(
                bento_repository_name=bento_repository.name, version=version
            )
        with io.BytesIO() as tar_io:
            bento_dir_path = bento.path
            if bento_dir_path is None:
                raise BentoMLException(f'Bento "{bento}" path cannot be None')
            with self.spin(text=f'Creating tar archive for Bento "{bento.tag}"..'):
                with tarfile.open(fileobj=tar_io, mode="w:gz") as tar:

                    def filter_(
                        tar_info: tarfile.TarInfo,
                    ) -> t.Optional[tarfile.TarInfo]:
                        if tar_info.path == "./models" or tar_info.path.startswith(
                            "./models/"
                        ):
                            return None
                        return tar_info

                    tar.add(bento_dir_path, arcname="./", filter=filter_)
            tar_io.seek(0, 0)
            with self.spin(text=f'Start uploading Bento "{bento.tag}"..'):
                yatai_rest_client.start_upload_bento(
                    bento_repository_name=bento_repository.name, version=version
                )

            file_size = tar_io.getbuffer().nbytes

            self.transmission_progress.update(
                upload_task_id, completed=0, total=file_size, visible=True
            )
            self.transmission_progress.start_task(upload_task_id)

            def io_cb(x: int):
                self.transmission_progress.update(upload_task_id, advance=x)

            wrapped_file = CallbackIOWrapper(
                io_cb,
                tar_io,
                "read",
            )
            finish_req = FinishUploadBentoSchema(
                status=BentoUploadStatus.SUCCESS,
                reason="",
            )
            try:
                resp = requests.put(
                    remote_bento.presigned_upload_url, data=wrapped_file
                )
                if resp.status_code != 200:
                    finish_req = FinishUploadBentoSchema(
                        status=BentoUploadStatus.FAILED,
                        reason=resp.text,
                    )
            except Exception as e:  # pylint: disable=broad-except
                finish_req = FinishUploadBentoSchema(
                    status=BentoUploadStatus.FAILED,
                    reason=str(e),
                )
            if finish_req.status is BentoUploadStatus.FAILED:
                self.log_progress.add_task(
                    f'[bold red]Failed to upload Bento "{bento.tag}"'
                )
            with self.spin(text="Submitting upload status to Yatai"):
                yatai_rest_client.finish_upload_bento(
                    bento_repository_name=bento_repository.name,
                    version=version,
                    req=finish_req,
                )
            if finish_req.status != BentoUploadStatus.SUCCESS:
                self.log_progress.add_task(
                    f'[bold red]Failed pushing Bento "{bento.tag}": {finish_req.reason}'
                )
            else:
                self.log_progress.add_task(
                    f'[bold green]Successfully pushed Bento "{bento.tag}"'
                )

    @inject
    def pull_bento(
        self,
        tag: t.Union[str, Tag],
        *,
        force: bool = False,
        bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ) -> "Bento":
        with Live(self.progress_group):
            download_task_id = self.transmission_progress.add_task(
                f'Pulling bento "{tag}"', start=False, visible=False
            )
            return self._do_pull_bento(
                tag,
                download_task_id,
                force=force,
                bento_store=bento_store,
                model_store=model_store,
            )

    @inject
    def _do_pull_bento(
        self,
        tag: t.Union[str, Tag],
        download_task_id: TaskID,
        *,
        force: bool = False,
        bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ) -> "Bento":
        try:
            bento = bento_store.get(tag)
            if not force:
                self.log_progress.add_task(
                    f'[bold blue]Bento "{tag}" exists in local model store'
                )
                return bento
            bento_store.delete(tag)
        except NotFound:
            pass
        _tag = Tag.from_taglike(tag)
        name = _tag.name
        version = _tag.version
        if version is None:
            raise BentoMLException(f'Bento "{_tag}" version can not be None')
        yatai_rest_client = get_current_yatai_rest_api_client()
        with self.spin(text=f'Fetching bento "{_tag}"'):
            remote_bento = yatai_rest_client.get_bento(
                bento_repository_name=name, version=version
            )
        if not remote_bento:
            raise BentoMLException(f'Bento "{_tag}" not found on Yatai')
        with ThreadPoolExecutor(
            max_workers=max(len(remote_bento.manifest.models), 1)
        ) as executor:

            def pull_model(model_tag: Tag):
                model_download_task_id = self.transmission_progress.add_task(
                    f'Pulling model "{model_tag}"', start=False, visible=False
                )
                self._do_pull_model(
                    model_tag,
                    model_download_task_id,
                    force=force,
                    model_store=model_store,
                )

            futures = executor.map(pull_model, remote_bento.manifest.models)
            list(futures)
        with self.spin(text=f'Getting a presigned download url for bento "{_tag}"'):
            remote_bento = yatai_rest_client.presign_bento_download_url(name, version)
        url = remote_bento.presigned_download_url
        response = requests.get(url, stream=True)
        if response.status_code != 200:
            raise BentoMLException(
                f'Failed to download bento "{_tag}": {response.text}'
            )
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024  # 1 Kibibyte
        with NamedTemporaryFile() as tar_file:
            self.transmission_progress.update(
                download_task_id, completed=0, total=total_size_in_bytes, visible=True
            )
            self.transmission_progress.start_task(download_task_id)
            for data in response.iter_content(block_size):
                self.transmission_progress.update(download_task_id, advance=len(data))
                tar_file.write(data)
            self.log_progress.add_task(
                f'[bold green]Finished downloading all bento "{_tag}" files'
            )
            tar_file.seek(0, 0)
            tar = tarfile.open(fileobj=tar_file, mode="r:gz")
            with fs.open_fs("temp://") as temp_fs:
                for member in tar.getmembers():
                    f = tar.extractfile(member)
                    if f is None:
                        continue
                    p = Path(member.name)
                    if p.parent != Path("."):
                        temp_fs.makedirs(str(p.parent), recreate=True)
                    temp_fs.writebytes(member.name, f.read())
                bento = Bento.from_fs(temp_fs)
                for model_tag in remote_bento.manifest.models:
                    with self.spin(text=f'Copying model "{model_tag}" to bento'):
                        copy_model(
                            model_tag,
                            src_model_store=model_store,
                            target_model_store=bento._model_store,  # type: ignore
                        )
                bento = bento.save(bento_store)
                self.log_progress.add_task(
                    f'[bold green]Successfully pulled bento "{_tag}"'
                )
                return bento

    def push_model(self, model: "Model", *, force: bool = False):
        with Live(self.progress_group):
            upload_task_id = self.transmission_progress.add_task(
                f'Pushing model "{model.tag}"', start=False, visible=False
            )
            self._do_push_model(model, upload_task_id, force=force)

    def _do_push_model(
        self, model: "Model", upload_task_id: TaskID, *, force: bool = False
    ):
        yatai_rest_client = get_current_yatai_rest_api_client()
        name = model.tag.name
        version = model.tag.version
        if version is None:
            raise BentoMLException(f'Model "{model.tag}" version cannot be None')
        info = model.info
        with self.spin(text=f'Fetching model repository "{name}"'):
            model_repository = yatai_rest_client.get_model_repository(
                model_repository_name=name
            )
        if not model_repository:
            with self.spin(text=f'Model repository "{name}" not found, creating now..'):
                model_repository = yatai_rest_client.create_model_repository(
                    req=CreateModelRepositorySchema(name=name, description="")
                )
        with self.spin(text=f'Try fetching model "{model.tag}" from Yatai..'):
            remote_model = yatai_rest_client.get_model(
                model_repository_name=name, version=version
            )
        if (
            not force
            and remote_model
            and remote_model.upload_status == ModelUploadStatus.SUCCESS
        ):
            self.log_progress.add_task(
                f'[bold blue]Model "{model.tag}" already exists in Yatai, skipping'
            )
            return
        if not remote_model:
            labels: t.List[LabelItemSchema] = [
                LabelItemSchema(key=key, value=value)
                for key, value in info.labels.items()
            ]
            with self.spin(text=f'Registering model "{model.tag}" with Yatai..'):
                yatai_rest_client.create_model(
                    model_repository_name=model_repository.name,
                    req=CreateModelSchema(
                        description="",
                        version=version,
                        build_at=info.creation_time,
                        manifest=ModelManifestSchema(
                            module=info.module,
                            metadata=info.metadata,
                            context=info.context.to_dict(),
                            options=dict(info.options),
                            api_version=info.api_version,
                            bentoml_version=info.context.bentoml_version,
                            size_bytes=calc_dir_size(model.path),
                        ),
                        labels=labels,
                    ),
                )
        with self.spin(
            text=f'Getting a presigned upload url for model "{model.tag}"..'
        ):
            remote_model = yatai_rest_client.presign_model_upload_url(
                model_repository_name=model_repository.name, version=version
            )
        with io.BytesIO() as tar_io:
            bento_dir_path = model.path
            with self.spin(text=f'Creating tar archive for model "{model.tag}"..'):
                with tarfile.open(fileobj=tar_io, mode="w:gz") as tar:
                    tar.add(bento_dir_path, arcname="./")
            tar_io.seek(0, 0)
            with self.spin(text=f'Start uploading model "{model.tag}"..'):
                yatai_rest_client.start_upload_model(
                    model_repository_name=model_repository.name, version=version
                )
            file_size = tar_io.getbuffer().nbytes
            self.transmission_progress.update(
                upload_task_id,
                description=f'Uploading model "{model.tag}"',
                total=file_size,
                visible=True,
            )
            self.transmission_progress.start_task(upload_task_id)

            def io_cb(x: int):
                self.transmission_progress.update(upload_task_id, advance=x)

            wrapped_file = CallbackIOWrapper(
                io_cb,
                tar_io,
                "read",
            )
            finish_req = FinishUploadModelSchema(
                status=ModelUploadStatus.SUCCESS,
                reason="",
            )
            try:
                resp = requests.put(
                    remote_model.presigned_upload_url, data=wrapped_file
                )
                if resp.status_code != 200:
                    finish_req = FinishUploadModelSchema(
                        status=ModelUploadStatus.FAILED,
                        reason=resp.text,
                    )
            except Exception as e:  # pylint: disable=broad-except
                finish_req = FinishUploadModelSchema(
                    status=ModelUploadStatus.FAILED,
                    reason=str(e),
                )
            if finish_req.status is ModelUploadStatus.FAILED:
                self.log_progress.add_task(
                    f'[bold red]Failed to upload model "{model.tag}"'
                )
            with self.spin(text="Submitting upload status to Yatai"):
                yatai_rest_client.finish_upload_model(
                    model_repository_name=model_repository.name,
                    version=version,
                    req=finish_req,
                )
            if finish_req.status != ModelUploadStatus.SUCCESS:
                self.log_progress.add_task(
                    f'[bold red]Failed pushing model "{model.tag}" : {finish_req.reason}'
                )
            else:
                self.log_progress.add_task(
                    f'[bold green]Successfully pushed model "{model.tag}"'
                )

    @inject
    def pull_model(
        self,
        tag: t.Union[str, Tag],
        *,
        force: bool = False,
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ) -> "Model":
        with Live(self.progress_group):
            download_task_id = self.transmission_progress.add_task(
                f'Pulling model "{tag}"', start=False, visible=False
            )
            return self._do_pull_model(
                tag, download_task_id, force=force, model_store=model_store
            )

    @inject
    def _do_pull_model(
        self,
        tag: t.Union[str, Tag],
        download_task_id: TaskID,
        *,
        force: bool = False,
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ) -> "Model":
        try:
            model = model_store.get(tag)
            if not force:
                self.log_progress.add_task(
                    f'[bold blue]Model "{tag}" already exists locally, skipping'
                )
                return model
            else:
                model_store.delete(tag)
        except NotFound:
            pass
        yatai_rest_client = get_current_yatai_rest_api_client()
        _tag = Tag.from_taglike(tag)
        name = _tag.name
        version = _tag.version
        if version is None:
            raise BentoMLException(f'Model "{_tag}" version cannot be None')
        with self.spin(text=f'Getting a presigned download url for model "{_tag}"..'):
            remote_model = yatai_rest_client.presign_model_download_url(name, version)
        if not remote_model:
            raise BentoMLException(f'Model "{_tag}" not found on Yatai')
        url = remote_model.presigned_download_url
        response = requests.get(url, stream=True)
        if response.status_code != 200:
            raise BentoMLException(
                f'Failed to download model "{_tag}": {response.text}'
            )
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024  # 1 Kibibyte
        with NamedTemporaryFile() as tar_file:
            self.transmission_progress.update(
                download_task_id,
                description=f'Downloading model "{_tag}"',
                total=total_size_in_bytes,
                visible=True,
            )
            self.transmission_progress.start_task(download_task_id)
            for data in response.iter_content(block_size):
                self.transmission_progress.update(download_task_id, advance=len(data))
                tar_file.write(data)
            self.log_progress.add_task(
                f'[bold green]Finished downloading model "{_tag}" files'
            )
            tar_file.seek(0, 0)
            tar = tarfile.open(fileobj=tar_file, mode="r:gz")
            with fs.open_fs("temp://") as temp_fs:
                for member in tar.getmembers():
                    f = tar.extractfile(member)
                    if f is None:
                        continue
                    p = Path(member.name)
                    if p.parent != Path("."):
                        temp_fs.makedirs(str(p.parent), recreate=True)
                    temp_fs.writebytes(member.name, f.read())
                model = Model.from_fs(temp_fs).save(model_store)
                self.log_progress.add_task(
                    f'[bold green]Successfully pulled model "{_tag}"'
                )
                return model


yatai_client = YataiClient()



from __future__ import annotations

import typing as t
import logging
from typing import TYPE_CHECKING

import torch

import bentoml
from bentoml import Tag

from ..utils.pkg import get_pkg_version
from ...exceptions import NotFound
from ...exceptions import BentoMLException
from ...exceptions import MissingDependencyException
from ..models.model import Model
from ..models.model import ModelContext
from .common.pytorch import torch

if TYPE_CHECKING:
    from ..models.model import ModelSignaturesType


_PL_IMPORT_ERROR = f"""\
`torch` is required in order to use module `{__name__}`\n
Refers to https://pytorch.org/get-started/locally/ to setup PyTorch correctly.
Then run `pip install torch`
"""


try:
    import torch
except ImportError:  # pragma: no cover
    raise MissingDependencyException(_PL_IMPORT_ERROR)

logger = logging.getLogger(__name__)
MODULE_NAME = "bentoml.torchscript"
MODEL_FILENAME = "savd_model.pt"


def get(tag_like: str | Tag) -> Model:
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    return model


def load_model(
    bentoml_model: str | Tag | Model,
    device_id: t.Optional[str] = "cpu",
) -> torch.ScriptModule:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        device_id (:code:`str`, `optional`):
            Optional devices to put the given model on. Refers to https://pytorch.org/docs/stable/tensor_attributes.html#torch.torch.device
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`torch.ScriptModule`: an instance of :obj:`torch.ScriptModule` from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml
        lit = bentoml.torchscript.load_model('lit_classifier:latest', device_id="cuda:0")
    """
    if isinstance(bentoml_model, (str, Tag)):
        bentoml_model = get(bentoml_model)

    if bentoml_model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {bentoml_model} was saved with module {bentoml_model.info.module}, failed loading with {MODULE_NAME}."
        )
    weight_file = bentoml_model.path_of(MODEL_FILENAME)
    model: torch.ScriptModule = torch.jit.load(weight_file, map_location=device_id)  # type: ignore[reportPrivateImportUsage]
    return model


def save_model(
    name: str,
    model: torch.ScriptModule,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: t.Dict[str, str] | None = None,
    custom_objects: t.Dict[str, t.Any] | None = None,
    metadata: t.Dict[str, t.Any] | None = None,
    _include_pytorch_lightning_version: bool = False,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`torch.ScriptModule`):
            Instance of model to be saved
        signatures (:code:`dict`, `optional`):
            A dictionary of method names and their corresponding signatures.
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`, default to :code:`None`):
            Custom metadata for given model.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import bentoml
        import torch

        TODO(jiang)
    """

    if _include_pytorch_lightning_version:
        framework_versions = {
            "torch": get_pkg_version("torch"),
            "pytorch_lightning": get_pkg_version("pytorch_lightning"),
        }
    else:
        framework_versions = {"torch": get_pkg_version("torch")}

    context: ModelContext = ModelContext(
        framework_name="torchscript",
        framework_versions=framework_versions,
    )

    if signatures is None:
        signatures = {"__call__": {"batchable": False}}
        logger.info(
            f"Using the default model signature ({signatures}) for model {name}."
        )

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        labels=labels,
        signatures=signatures,
        custom_objects=custom_objects,
        options=None,
        context=context,
        metadata=metadata,
    ) as _model:
        weight_file = _model.path_of(MODEL_FILENAME)
        torch.jit.save(model, weight_file)  # type: ignore
        return _model.tag


def get_runnable(bento_model: Model):
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """
    from .common.pytorch import partial_class
    from .common.pytorch import PytorchModelRunnable
    from .common.pytorch import make_pytorch_runnable_method

    for method_name, options in bento_model.info.signatures.items():
        PytorchModelRunnable.add_method(
            make_pytorch_runnable_method(method_name),
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )
    return partial_class(
        PytorchModelRunnable,
        bento_model=bento_model,
        loader=load_model,
    )



import os
import typing as t
from typing import TYPE_CHECKING

import numpy as np
from simple_di import inject
from simple_di import Provide

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import PTH_EXT
from ..models import YAML_EXT
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..runner.utils import Params
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from .. import external_typing as ext
    from ..models import ModelStore
try:
    # pylint: disable=unused-import
    import torch
    import detectron2  # noqa F401
    import detectron2.config as config
    import detectron2.modeling as modeling
    import detectron2.checkpoint as checkpoint
    from detectron2.checkpoint import DetectionCheckpointer
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """detectron2 is required in order to use module `bentoml.detectron`,
        install detectron2 with `pip install detectron2`. For more
        information, refers to
        https://detectron2.readthedocs.io/en/latest/tutorials/install.html
        """
    )

MODULE_NAME = "bentoml.detectron"


@inject
def load(
    tag: t.Union[str, Tag],
    device: str = "cpu",
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "torch.nn.Module":
    """
    Load a model from BentoML local modelstore with given tag.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        device (:code:`str`, `optional`, default to :code:`cpu`):
            Device type to cast model. Default behaviour similar to :obj:`torch.device("cuda")` Options: "cuda" or "cpu". If None is specified then return default config.MODEL.DEVICE
        model_store (`~bentoml._internal.models.ModelStore`, default to :code:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`torch.nn.Module`: an instance of `torch.nn.Module`

    Examples:

    .. code-block:: python

        import bentoml
        model = bentoml.detectron.load("my_detectron_model")
    """  # noqa: LN001

    model_info = model_store.get(tag)
    if model_info.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model_info.info.module}, failed loading with {MODULE_NAME}."
        )

    cfg: config.CfgNode = config.get_cfg()

    weight_path = model_info.path_of(f"{SAVE_NAMESPACE}{PTH_EXT}")
    yaml_path = model_info.path_of(f"{SAVE_NAMESPACE}{YAML_EXT}")

    if os.path.isfile(yaml_path):
        cfg.merge_from_file(yaml_path)

    if device:
        cfg.MODEL.DEVICE = device

    model: "torch.nn.Module" = modeling.build_model(cfg)
    if device:
        model.to(device)

    model.eval()

    checkpointer: "DetectionCheckpointer" = checkpoint.DetectionCheckpointer(model)

    checkpointer.load(weight_path)
    return model


def save(
    name: str,
    model: "torch.nn.Module",
    *,
    model_config: t.Optional[config.CfgNode] = None,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`torch.nn.Module`):
            Instance of detectron2 model to be saved.
        model_config (`detectron2.config.CfgNode`, `optional`, default to :code:`None`):
            model config from :meth:`detectron2.model_zoo.get_config_file`
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.ModelStore`, default to :code:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import bentoml

        # import some common detectron2 utilities
        import detectron2
        from detectron2 import model_zoo
        from detectron2.config import get_cfg
        from detectron2.modeling import build_model

        model_url: str = "COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"
        cfg: "CfgNode" = get_cfg()
        cfg.merge_from_file(model_zoo.get_config_file(model_url))
        # set threshold for this model
        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5
        cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(model_url)
        cloned = cfg.clone()
        cloned.MODEL.DEVICE = "cpu"
        model: torch.nn.Module = build_model(cloned)

        tag = bentoml.detectron.save(
            "my_detectron_model",
            model,
            model_config=cfg,
        )

        # load the model back:
        loaded = bentoml.detectron.load("my_detectron_model:latest")
        # or:
        loaded = bentoml.detectron.load(tag)

    """  # noqa

    context: t.Dict[str, t.Any] = {
        "framework_name": "detectron2",
        "pip_dependencies": [
            f"detectron2=={get_pkg_version('detectron2')}",
            f"torch=={get_pkg_version('torch')}",
        ],
    }
    options: t.Dict[str, t.Any] = dict()

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        labels=labels,
        custom_objects=custom_objects,
        options=options,
        context=context,
        metadata=metadata,
    ) as _model:
        checkpointer = checkpoint.DetectionCheckpointer(model, save_dir=_model.path)
        checkpointer.save(SAVE_NAMESPACE)
        if model_config:
            with open(
                _model.path_of(f"{SAVE_NAMESPACE}{YAML_EXT}"),
                "w",
                encoding="utf-8",
            ) as ouf:
                ouf.write(model_config.dump())

        return _model.tag


from .common.model_runner import BaseModelRunner


class _DetectronRunner(BaseModelRunner):
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        name: t.Optional[str] = None,
    ):
        super().__init__(tag, name=name)
        self._predict_fn_name = predict_fn_name

    @property
    def num_replica(self) -> int:
        if self.resource_quota.nvidia_gpu:
            return self.resource_quota.nvidia_gpu
        return 1

    @property
    def _device(self) -> str:
        if self.resource_quota.nvidia_gpu:
            return "cuda"
        return "cpu"

    def _setup(self) -> None:
        self._model = load(self._tag, self._device, model_store=self.model_store)
        self._predict_fn = getattr(self._model, self._predict_fn_name)

    def _run_batch(  # type: ignore
        self,
        *args: t.Union["ext.NpNDArray", torch.Tensor],
    ) -> "ext.NpNDArray":
        params = Params[t.Union["ext.NpNDArray", torch.Tensor]](*args)

        def _mapping(item: t.Union["ext.NpNDArray", torch.Tensor]) -> torch.Tensor:
            if isinstance(item, np.ndarray):
                return torch.Tensor(item, device=self._device)
            return item

        params = params.map(_mapping)

        inputs = [{"image": image} for image in params.args]

        res: "torch.Tensor" = self._predict_fn(inputs)
        return np.asarray(res)  # type: ignore


def load_runner(
    tag: t.Union[str, Tag],
    predict_fn_name: str = "__call__",
    *,
    name: t.Optional[str] = None,
) -> _DetectronRunner:
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. :func:`bentoml.detectron.load_runner` implements a Runner class that
    wrap around a :obj:`torch.nn.Module` model, which optimize it for the BentoML runtime.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        predict_fn_name (:code:`str`, default to :code:`__call__`):
            Options for inference functions. Default to `__call__`
        model_store (`~bentoml._internal.models.ModelStore`, default to :code:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml._internal.runner.Runner`: Runner instances for :mod:`bentoml.detectron` model

    Examples:

    .. code-block:: python

        import bentoml
        import numpy as np

        runner = bentoml.detectron.load_runner(tag)
        runner.run_batch(np.array([[1,2,3,]]))
    """
    return _DetectronRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        name=name,
    )



from __future__ import annotations

import typing as t
import logging
from typing import TYPE_CHECKING

import attr

from bentoml.exceptions import BentoMLException

from ..tag import Tag
from ..models import Model
from ..runner import Runner
from ..bento.bento import get_default_bento_readme
from .inference_api import InferenceAPI
from ..io_descriptors import IODescriptor

if TYPE_CHECKING:
    from .. import external_typing as ext
    from ..bento import Bento

    WSGI_APP = t.Callable[
        [t.Callable[..., t.Any], t.Mapping[str, t.Any]], t.Iterable[bytes]
    ]

logger = logging.getLogger(__name__)


def add_inference_api(
    svc: Service,
    func: t.Callable[..., t.Any],
    input: IODescriptor[t.Any],  # pylint: disable=redefined-builtin
    output: IODescriptor[t.Any],
    name: t.Optional[str],
    doc: t.Optional[str],
    route: t.Optional[str],
) -> None:
    api = InferenceAPI(
        name=name if name is not None else func.__name__,
        user_defined_callback=func,
        input_descriptor=input,
        output_descriptor=output,
        doc=doc,
        route=route,
    )

    if api.name in svc.apis:
        raise BentoMLException(
            f"API {api.name} is already defined in Service {svc.name}"
        )

    from ..io_descriptors import Multipart

    if isinstance(output, Multipart):
        logger.warning(
            f"Found Multipart as the output of API '{api.name}'. Multipart responses are rarely used in the real world, and few clients/browsers support it. Make sure you know what you are doing."
        )

    svc.apis[api.name] = api


def get_valid_service_name(user_provided_svc_name: str) -> str:
    lower_name = user_provided_svc_name.lower()

    if user_provided_svc_name != lower_name:
        logger.warning(
            f"converting {user_provided_svc_name} to lowercase: {lower_name}"
        )

    # Service name must be a valid Tag name; create a dummy tag to use its validation
    Tag(lower_name)
    return lower_name


@attr.define(frozen=True)
class Service:
    """The service definition is the manifestation of the Service Oriented Architecture
    and the core building block in BentoML where users define the service runtime
    architecture and model serving logic.

    A BentoML service is defined via instantiate this Service class. When creating a
    Service instance, user must provide a Service name and list of runners that are
    required by this Service. The instance can then be used to define InferenceAPIs via
    the `api` decorator.
    """

    name: str
    runners: t.List[Runner]
    models: t.List[Model]

    mount_apps: t.List[t.Tuple[ext.ASGIApp, str, str]] = attr.field(
        init=False, factory=list
    )
    middlewares: t.List[
        t.Tuple[t.Type[ext.AsgiMiddleware], t.Dict[str, t.Any]]
    ] = attr.field(init=False, factory=list)

    apis: t.Dict[str, InferenceAPI] = attr.field(init=False, factory=dict)

    # Tag/Bento are only set when the service was loaded from a bento
    tag: Tag | None = attr.field(init=False, default=None)
    bento: Bento | None = attr.field(init=False, default=None)

    # Working dir and Import path of the service, set when the service was imported
    _working_dir: str | None = attr.field(init=False, default=None)
    _import_str: str | None = attr.field(init=False, default=None)

    def __init__(
        self,
        name: str,
        *,
        runners: t.List[Runner] | None = None,
        models: t.List[Model] | None = None,
    ):
        """

        Args:
            name:
            runners:
            models:
        """
        name = get_valid_service_name(name)

        # validate runners list contains Runner instances and runner names are unique
        if runners is not None:
            runner_names: t.Set[str] = set()
            for r in runners:
                assert isinstance(
                    r, Runner
                ), f'Service runners list can only contain bentoml.Runner instances, type "{type(r)}" found.'

                if r.name in runner_names:
                    raise ValueError(
                        f"Found duplicate name `{r.name}` in service runners."
                    )
                runner_names.add(r.name)

        # validate models list contains Model instances
        if models is not None:
            for model in models:
                assert isinstance(
                    model, Model
                ), f'Service models list can only contain bentoml.Model instances, type "{type(model)}" found.'

        self.__attrs_init__(  # type: ignore
            name=name,
            runners=[] if runners is None else runners,
            models=[] if models is None else models,
        )

    def api(
        self,
        input: IODescriptor[t.Any],  # pylint: disable=redefined-builtin
        output: IODescriptor[t.Any],
        name: t.Optional[str] = None,
        doc: t.Optional[str] = None,
        route: t.Optional[str] = None,
    ) -> t.Callable[[t.Callable[..., t.Any]], t.Callable[..., t.Any]]:
        """Decorator for adding InferenceAPI to this service"""

        D = t.TypeVar("D", bound=t.Callable[..., t.Any])

        def decorator(func: D) -> D:
            add_inference_api(self, func, input, output, name, doc, route)
            return func

        return decorator

    def __str__(self):
        if self.bento:
            return f'bentoml.Service(tag="{self.tag}", ' f'path="{self.bento.path}")'
        elif self._import_str and self._working_dir:
            return (
                f'bentoml.Service(name="{self.name}", '
                f'import_str="{self._import_str}", '
                f'working_dir="{self._working_dir}")'
            )
        else:
            return (
                f'bentoml.Service(name="{self.name}", '
                f'runners=[{",".join([r.name for r in self.runners])}])'
            )

    def __repr__(self):
        return self.__str__()

    @property
    def doc(self) -> str:
        if self.bento is not None:
            return self.bento.doc

        return get_default_bento_readme(self)

    def openapi_doc(self):
        from .openapi import get_service_openapi_doc

        return get_service_openapi_doc(self)

    def on_asgi_app_startup(self) -> None:
        pass

    def on_asgi_app_shutdown(self) -> None:
        pass

    @property
    def asgi_app(self) -> "ext.ASGIApp":
        from ..server.service_app import ServiceAppFactory

        return ServiceAppFactory(self)()

    def mount_asgi_app(
        self, app: "ext.ASGIApp", path: str = "/", name: t.Optional[str] = None
    ) -> None:
        self.mount_apps.append((app, path, name))  # type: ignore

    def mount_wsgi_app(
        self, app: WSGI_APP, path: str = "/", name: t.Optional[str] = None
    ) -> None:
        # TODO: Migrate to a2wsgi
        from starlette.middleware.wsgi import WSGIMiddleware

        self.mount_apps.append((WSGIMiddleware(app), path, name))  # type: ignore

    def add_asgi_middleware(
        self, middleware_cls: t.Type["ext.AsgiMiddleware"], **options: t.Any
    ) -> None:
        self.middlewares.append((middleware_cls, options))


def on_load_bento(svc: Service, bento: Bento):
    object.__setattr__(svc, "bento", bento)
    object.__setattr__(svc, "tag", bento.info.tag)


def on_import_svc(svc: Service, working_dir: str, import_str: str):
    object.__setattr__(svc, "_working_dir", working_dir)
    object.__setattr__(svc, "_import_str", import_str)



from __future__ import annotations

import sys
import typing as t
import logging

import click

from bentoml.bentos import containerize as containerize_bento

from ..utils import kwargs_transformers
from ..utils.docker import validate_tag

logger = logging.getLogger("bentoml")


def containerize_transformer(
    value: t.Iterable[str] | str | bool | None,
) -> t.Iterable[str] | str | bool | None:
    if value is None:
        return
    if isinstance(value, tuple) and not value:
        return
    return value


def add_containerize_command(cli: click.Group) -> None:
    @cli.command()
    @click.argument("bento_tag", type=click.STRING)
    @click.option(
        "-t",
        "--docker-image-tag",
        help="Name and optionally a tag (format: 'name:tag'), defaults to bento tag.",
        required=False,
        callback=validate_tag,
    )
    @click.option(
        "--add-host",
        multiple=True,
        help="Add a custom host-to-IP mapping (format: 'host:ip').",
    )
    @click.option(
        "--allow",
        multiple=True,
        default=None,
        help="Allow extra privileged entitlement (e.g., 'network.host', 'security.insecure').",
    )
    @click.option("--build-arg", multiple=True, help="Set build-time variables.")
    @click.option(
        "--build-context",
        multiple=True,
        help="Additional build contexts (e.g., name=path).",
    )
    @click.option(
        "--builder",
        type=click.STRING,
        default=None,
        help="Override the configured builder instance.",
    )
    @click.option(
        "--cache-from",
        multiple=True,
        default=None,
        help="External cache sources (e.g., 'user/app:cache', 'type=local,src=path/to/dir').",
    )
    @click.option(
        "--cache-to",
        multiple=True,
        default=None,
        help="Cache export destinations (e.g., 'user/app:cache', 'type=local,dest=path/to/dir').",
    )
    @click.option(
        "--cgroup-parent",
        type=click.STRING,
        default=None,
        help="Optional parent cgroup for the container.",
    )
    @click.option(
        "--iidfile",
        type=click.STRING,
        default=None,
        help="Write the image ID to the file.",
    )
    @click.option("--label", multiple=True, help="Set metadata for an image.")
    @click.option(
        "--load",
        is_flag=True,
        default=False,
        help="Shorthand for '--output=type=docker'.",
    )
    @click.option(
        "--metadata-file",
        type=click.STRING,
        default=None,
        help="Write build result metadata to the file.",
    )
    @click.option(
        "--network",
        type=click.STRING,
        default=None,
        help="Set the networking mode for the 'RUN' instructions during build (default 'default').",
    )
    @click.option(
        "--no-cache",
        is_flag=True,
        default=False,
        help="Do not use cache when building the image.",
    )
    @click.option(
        "--no-cache-filter",
        multiple=True,
        help="Do not cache specified stages.",
    )
    @click.option(
        "--output",
        multiple=True,
        default=None,
        help="Output destination (format: 'type=local,dest=path').",
    )
    @click.option(
        "--platform", default=None, multiple=True, help="Set target platform for build."
    )
    @click.option(
        "--progress",
        default="auto",
        type=click.Choice(["auto", "tty", "plain"]),
        help="Set type of progress output ('auto', 'plain', 'tty'). Use plain to show container output.",
    )
    @click.option(
        "--pull",
        is_flag=True,
        default=False,
        help="Always attempt to pull all referenced images.",
    )
    @click.option(
        "--push",
        is_flag=True,
        default=False,
        help="Shorthand for '--output=type=registry'.",
    )
    @click.option(
        "--secret",
        multiple=True,
        default=None,
        help="Secret to expose to the build (format: 'id=mysecret[,src=/local/secret]').",
    )
    @click.option("--shm-size", default=None, help="Size of '/dev/shm'.")
    @click.option(
        "--ssh",
        type=click.STRING,
        default=None,
        help="SSH agent socket or keys to expose to the build (format: 'default|<id>[=<socket>|<key>[,<key>]]').",
    )
    @click.option(
        "--target",
        type=click.STRING,
        default=None,
        help="Set the target build stage to build.",
    )
    @click.option(
        "--ulimit", type=click.STRING, default=None, help="Ulimit options (default [])."
    )
    @kwargs_transformers(transformer=containerize_transformer)
    def containerize(  # type: ignore
        bento_tag: str,
        docker_image_tag: str,
        add_host: t.Iterable[str],
        allow: t.Iterable[str],
        build_arg: t.List[str],
        build_context: t.List[str],
        builder: str,
        cache_from: t.List[str],
        cache_to: t.List[str],
        cgroup_parent: str,
        iidfile: str,
        label: t.List[str],
        load: bool,
        network: str,
        metadata_file: str,
        no_cache: bool,
        no_cache_filter: t.List[str],
        output: t.List[str],
        platform: t.List[str],
        progress: t.Literal["auto", "tty", "plain"],
        pull: bool,
        push: bool,
        secret: t.List[str],
        shm_size: str,
        ssh: str,
        target: str,
        ulimit: str,
    ) -> None:
        """Containerizes given Bento into a ready-to-use Docker image.

        \b
        BENTO is the target BentoService to be containerized, referenced by its name
        and version in format of name:version. For example: "iris_classifier:v1.2.0"

        `bentoml containerize` command also supports the use of the `latest` tag
        which will automatically use the last built version of your Bento.

        You can provide a tag for the image built by Bento using the
        `--tag` flag. Additionally, you can provide a `--push` flag,
        which will push the built image to the Docker repository specified by the
        image tag.

        You can also prefixing the tag with a hostname for the repository you wish
        to push to.
        e.g. `bentoml containerize IrisClassifier:latest --push --tag
        repo-address.com:username/iris` would build a Docker image called
        `username/iris:latest` and push that to docker repository at repo-address.com.

        By default, the `containerize` command will use the current credentials
        provided by Docker daemon.

        `bentoml containerize` also uses Docker Buildx as backend, in place for normal `docker build`.
        By doing so, BentoML will leverage Docker Buildx features such as multi-node
        builds for cross-platform images, Full BuildKit capabilities with all of the
        familiar UI from `docker build`.

        We also pass all given args for `docker buildx` through `bentoml containerize` with ease.
        """
        add_hosts = {}
        if add_host:
            for host in add_host:
                host_name, ip = host.split(":")
                add_hosts[host_name] = ip

        allow_ = []
        if allow:
            allow_ = list(allow)

        build_args = {}
        if build_arg:
            for build_arg_str in build_arg:
                key, value = build_arg_str.split("=")
                build_args[key] = value

        build_context_ = {}
        if build_context:
            for build_context_str in build_context:
                key, value = build_context_str.split("=")
                build_context_[key] = value

        labels = {}
        if label:
            for label_str in label:
                key, value = label_str.split("=")
                labels[key] = value

        if output:
            output_ = {}
            for arg in output:
                if "," in arg:
                    for val in arg.split(","):
                        k, v = val.split("=")
                        output_[k] = v
                key, value = arg.split("=")
                output_[key] = value
        else:
            output_ = None

        if not platform:
            load = True
        else:
            if len(platform) > 1:
                logger.warning(
                    "Multiple `--platform` arguments were found. Make sure to also use `--push` to push images to a repository or generated images will not be saved. For more information, see https://docs.docker.com/engine/reference/commandline/buildx_build/#load."
                )
            else:
                load = True

        exit_code = not containerize_bento(
            bento_tag,
            docker_image_tag=docker_image_tag,
            add_host=add_hosts,
            allow=allow_,
            build_args=build_args,
            build_context=build_context_,
            builder=builder,
            cache_from=cache_from,
            cache_to=cache_to,
            cgroup_parent=cgroup_parent,
            iidfile=iidfile,
            labels=labels,
            load=load,
            metadata_file=metadata_file,
            network=network,
            no_cache=no_cache,
            no_cache_filter=no_cache_filter,
            output=output_,  # type: ignore
            platform=platform,
            progress=progress,
            pull=pull,
            push=push,
            quiet=logger.getEffectiveLevel() == logging.ERROR,
            secrets=secret,
            shm_size=shm_size,
            ssh=ssh,
            target=target,
            ulimit=ulimit,
        )
        sys.exit(exit_code)



from __future__ import annotations

import json
import typing as t
import dataclasses
from typing import TYPE_CHECKING

from starlette.requests import Request
from starlette.responses import Response

from .base import IODescriptor
from ..types import LazyType
from ..utils.http import set_cookies
from ...exceptions import BadInput
from ...exceptions import MissingDependencyException

if TYPE_CHECKING:
    from types import UnionType

    import pydantic

    from .. import external_typing as ext
    from ..context import InferenceApiContext as Context

    _SerializableObj: t.TypeAlias = t.Union[
        "ext.NpNDArray",
        "ext.PdDataFrame",
        t.Type["pydantic.BaseModel"],
        t.Any,
    ]


JSONType = t.Union[str, t.Dict[str, t.Any], "pydantic.BaseModel"]

MIME_TYPE_JSON = "application/json"


class DefaultJsonEncoder(json.JSONEncoder):  # pragma: no cover
    def default(self, o: _SerializableObj) -> t.Any:
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)

        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(o):
            return o.tolist()
        if LazyType["ext.NpGeneric"]("numpy.generic").isinstance(o):
            return o.item()
        if LazyType["ext.PdDataFrame"]("pandas.DataFrame").isinstance(o):
            return o.to_dict()  # type: ignore
        if LazyType["ext.PdSeries"]("pandas.Series").isinstance(o):
            return o.to_dict()  # type: ignore
        if LazyType["pydantic.BaseModel"]("pydantic.BaseModel").isinstance(o):
            obj_dict = o.dict()
            if "__root__" in obj_dict:
                obj_dict = obj_dict.get("__root__")
            return obj_dict

        return super().default(o)


class JSON(IODescriptor[JSONType]):
    """
    :code:`JSON` defines API specification for the inputs/outputs of a Service, where either
    inputs will be converted to or outputs will be converted from a JSON representation
    as specified in your API function signature.

    Sample implementation of a sklearn service:

    .. code-block:: python

        # sklearn_svc.py
        import pandas as pd
        import numpy as np
        from bentoml.io import PandasDataFrame, JSON
        import bentoml.sklearn

        input_spec = PandasDataFrame.from_sample(pd.DataFrame(np.array([[5,4,3,2]])))

        runner = bentoml.sklearn.load_runner("sklearn_model_clf")

        svc = bentoml.Service("iris-classifier", runners=[runner])

        @svc.api(input=input_spec, output=JSON())
        def predict(input_arr: pd.DataFrame):
            res = runner.run_batch(input_arr)  # type: np.ndarray
            return {"res":pd.DataFrame(res).to_json(orient='record')}

    Users then can then serve this service with :code:`bentoml serve`:

    .. code-block:: bash

        % bentoml serve ./sklearn_svc.py:svc --auto-reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "iris-classifier" defined in "sklearn_svc.py"
        [INFO] API Server running on http://0.0.0.0:3000

    Users can then send requests to the newly started services with any client:

    .. tabs::

        .. code-tab:: python

            import requests
            requests.post(
                "http://0.0.0.0:3000/predict",
                headers={"content-type": "application/json"},
                data='[{"0":5,"1":4,"2":3,"3":2}]'
            ).text

        .. code-tab:: bash

            % curl -X POST -H "Content-Type: application/json" --data '[{"0":5,"1":4,"2":3,"3":2}]' http://0.0.0.0:3000/predict

            {"res":"[{\"0\":1}]"}%

    Args:
        pydantic_model (:code:`pydantic.BaseModel`, `optional`, default to :code:`None`):
            Pydantic model schema.
        validate_json (:code:`bool`, `optional`, default to :code:`True`): If True, then use
            Pydantic model specified above to validate given JSON.
        json_encoder (:code:`Type[json.JSONEncoder]`, default to :code:`~bentoml._internal.io_descriptor.json.DefaultJsonEncoder`):
            JSON encoder class.

    Returns:
        :obj:`~bentoml._internal.io_descriptors.IODescriptor`: IO Descriptor that in JSON format.
    """

    def __init__(
        self,
        pydantic_model: t.Type[pydantic.BaseModel] | None = None,
        validate_json: bool = True,
        json_encoder: t.Type[json.JSONEncoder] = DefaultJsonEncoder,
    ):
        if pydantic_model is not None:
            try:
                import pydantic
            except ImportError:
                raise MissingDependencyException(
                    "`pydantic` must be installed to use `pydantic_model`"
                )
            assert issubclass(
                pydantic_model, pydantic.BaseModel
            ), "`pydantic_model` must be a subclass of `pydantic.BaseModel`"

        self._pydantic_model = pydantic_model
        self._validate_json = validate_json
        self._json_encoder = json_encoder

    def input_type(self) -> "UnionType":
        return JSONType

    def openapi_schema_type(self) -> t.Dict[str, t.Any]:
        if self._pydantic_model is None:
            return {"type": "object"}
        return self._pydantic_model.schema()

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {MIME_TYPE_JSON: {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {MIME_TYPE_JSON: {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> JSONType:
        json_str = await request.body()
        if self._pydantic_model is not None and self._validate_json:
            try:
                import pydantic
            except ImportError:
                raise MissingDependencyException(
                    "`pydantic` must be installed to use `pydantic_model`"
                ) from None
            try:
                pydantic_model = self._pydantic_model.parse_raw(json_str)
                return pydantic_model.dict()
            except pydantic.ValidationError as e:
                raise BadInput(f"Json validation error: {e}") from None
        else:
            try:
                return json.loads(json_str)
            except json.JSONDecodeError as e:
                raise BadInput(f"Json validation error: {e}") from None

    async def to_http_response(self, obj: JSONType, ctx: Context | None = None):
        json_str = json.dumps(
            obj,
            cls=self._json_encoder,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
        )

        if ctx is not None:
            res = Response(
                json_str,
                media_type=MIME_TYPE_JSON,
                headers=ctx.response.metadata,  # type: ignore (bad starlette types)
                status_code=ctx.response.status_code,
            )
            set_cookies(res, ctx.response.cookies)
            return res
        else:
            return Response(json_str, media_type=MIME_TYPE_JSON)



import time
import typing as t
import asyncio
import logging
import functools
import traceback
import collections

import numpy as np

from ..utils import cached_property
from ..utils.alg import TokenBucket

logger = logging.getLogger(__name__)


class NonBlockSema:
    def __init__(self, count):
        self.sema = count

    def acquire(self):
        if self.sema < 1:
            return False
        self.sema -= 1
        return True

    def is_locked(self):
        return self.sema < 1

    def release(self):
        self.sema += 1


class Optimizer:
    """
    Analyse historical data to optimize CorkDispatcher.
    """

    N_KEPT_SAMPLE = 50  # amount of outbound info kept for inferring params
    N_SKIPPED_SAMPLE = 2  # amount of outbound info skipped after init
    INTERVAL_REFRESH_PARAMS = 5  # seconds between each params refreshing

    def __init__(self):
        """
        assume the outbound duration follows duration = o_a * n + o_b
        (all in seconds)
        """
        self.o_stat = collections.deque(
            maxlen=self.N_KEPT_SAMPLE
        )  # to store outbound stat data
        self.o_a = 2
        self.o_b = 1

        self.wait = 0.01  # the avg wait time before outbound called

        self._refresh_tb = TokenBucket(2)  # to limit params refresh interval
        self._outbound_counter = 0

    def log_outbound(self, n, wait, duration):
        if (
            self._outbound_counter <= self.N_SKIPPED_SAMPLE
        ):  # skip inaccurate info at beginning
            self._outbound_counter += 1
            return

        self.o_stat.append((n, duration, wait))

        if self._refresh_tb.consume(1, 1.0 / self.INTERVAL_REFRESH_PARAMS, 1):
            self.trigger_refresh()

    def trigger_refresh(self):
        x = tuple((i, 1) for i, _, _ in self.o_stat)
        y = tuple(i for _, i, _ in self.o_stat)
        _o_a, _o_b = np.linalg.lstsq(x, y, rcond=None)[0]
        _o_w = sum(w for _, _, w in self.o_stat) * 1.0 / len(self.o_stat)

        self.o_a, self.o_b = max(0.000001, _o_a), max(0, _o_b)
        self.wait = max(0, _o_w)
        logger.debug(
            "Dynamic batching optimizer params updated: o_a: %.6f, o_b: %.6f, wait: %.6f",
            _o_a,
            _o_b,
            _o_w,
        )


T_IN = t.TypeVar("T_IN")
T_OUT = t.TypeVar("T_OUT")


class CorkDispatcher:
    """
    A decorator that:
        * wrap batch function
        * implement CORK algorithm to cork & release calling of wrapped function
    The wrapped function should be an async function.
    """

    def __init__(
        self,
        max_latency_in_ms: int,
        max_batch_size: int,
        shared_sema: t.Optional[NonBlockSema] = None,
        fallback: t.Optional[t.Callable[[], t.Any]] = None,
    ):
        """
        params:
            * max_latency_in_ms: max_latency_in_ms for inbound tasks in milliseconds
            * max_batch_size: max batch size of inbound tasks
            * shared_sema: semaphore to limit concurrent outbound tasks
            * fallback: callable to return fallback result
        raises:
            * all possible exceptions the decorated function has
        """
        self.max_latency_in_ms = max_latency_in_ms / 1000.0
        self.callback = None
        self.fallback = fallback
        self.optimizer = Optimizer()
        self.max_batch_size = int(max_batch_size)
        self.tick_interval = 0.001

        self._controller = None
        self._queue = collections.deque()  # TODO(hrmthw): maxlen
        self._sema = shared_sema if shared_sema else NonBlockSema(1)

    def shutdown(self):
        if self._controller is not None:
            self._controller.cancel()
        try:
            while True:
                _, _, fut = self._queue.pop()
                fut.cancel()
        except IndexError:
            pass

    @cached_property
    def _loop(self):
        return asyncio.get_event_loop()

    @cached_property
    def _wake_event(self):
        return asyncio.Condition()

    def __call__(
        self,
        callback: t.Callable[
            [t.Iterable[T_IN]], t.Coroutine[None, None, t.Iterable[T_OUT]]
        ],
    ) -> t.Callable[[T_IN], t.Coroutine[None, None, T_OUT]]:
        self.callback = callback

        @functools.wraps(callback)
        async def _func(data):
            if self._controller is None:
                self._controller = self._loop.create_task(self.controller())
            try:
                r = await self.inbound_call(data)
            except asyncio.CancelledError:
                return None if self.fallback is None else self.fallback()
            if isinstance(r, Exception):
                raise r
            return r

        return _func

    async def controller(self):
        """
        A standalone coroutine to wait/dispatch calling.
        """
        while True:
            try:
                async with self._wake_event:  # block until there's any request in queue
                    await self._wake_event.wait_for(self._queue.__len__)

                n = len(self._queue)
                dt = self.tick_interval
                decay = 0.95  # the decay rate of wait time
                now = time.time()
                w0 = now - self._queue[0][0]
                wn = now - self._queue[-1][0]
                a = self.optimizer.o_a
                b = self.optimizer.o_b

                if n > 1 and (w0 + a * n + b) >= self.max_latency_in_ms:
                    self._queue.popleft()[2].cancel()
                    continue
                if self._sema.is_locked():
                    if n == 1 and w0 >= self.max_latency_in_ms:
                        self._queue.popleft()[2].cancel()
                        continue
                    await asyncio.sleep(self.tick_interval)
                    continue
                if n * (wn + dt + (a or 0)) <= self.optimizer.wait * decay:
                    await asyncio.sleep(self.tick_interval)
                    continue

                n_call_out = min(
                    self.max_batch_size,
                    n,
                )
                # call
                self._sema.acquire()
                inputs_info = tuple(self._queue.pop() for _ in range(n_call_out))
                self._loop.create_task(self.outbound_call(inputs_info))
            except asyncio.CancelledError:
                break
            except Exception:  # pylint: disable=broad-except
                logger.error(traceback.format_exc())

    async def inbound_call(self, data):
        now = time.time()
        future = self._loop.create_future()
        input_info = (now, data, future)
        self._queue.append(input_info)
        async with self._wake_event:
            self._wake_event.notify_all()
        return await future

    async def outbound_call(self, inputs_info):
        _time_start = time.time()
        _done = False
        logger.debug("Dynamic batching cork released, batch size: %d", len(inputs_info))
        try:
            outputs = await self.callback(tuple(d for _, d, _ in inputs_info))
            assert len(outputs) == len(inputs_info)
            for (_, _, fut), out in zip(inputs_info, outputs):
                if not fut.done():
                    fut.set_result(out)
            _done = True
            self.optimizer.log_outbound(
                n=len(inputs_info),
                wait=_time_start - inputs_info[-1][0],
                duration=time.time() - _time_start,
            )
        except asyncio.CancelledError:
            pass
        except Exception as e:  # pylint: disable=broad-except
            for _, _, fut in inputs_info:
                if not fut.done():
                    fut.set_result(e)
            _done = True
        finally:
            if not _done:
                for _, _, fut in inputs_info:
                    if not fut.done():
                        fut.cancel()
            self._sema.release()



from __future__ import annotations

import typing as t
from typing import TYPE_CHECKING

from starlette.requests import Request
from multipart.multipart import parse_options_header
from starlette.responses import Response

from .base import IODescriptor
from ...exceptions import InvalidArgument
from ...exceptions import BentoMLException
from ..utils.formparser import populate_multipart_requests
from ..utils.formparser import concat_to_multipart_response

if TYPE_CHECKING:
    from types import UnionType

    from ..types import LazyType
    from ..context import InferenceApiContext as Context


class Multipart(IODescriptor[t.Any]):
    """
    :code:`Multipart` defines API specification for the inputs/outputs of a Service, where inputs/outputs
    of a Service can receive/send a *multipart* request/responses as specified in your API function signature.

    Sample implementation of a sklearn service:

    .. code-block:: python

        # sklearn_svc.py
        import bentoml
        from bentoml.io import NumpyNdarray, Multipart, JSON
        import bentoml.sklearn

        runner = bentoml.sklearn.load_runner("sklearn_model_clf")

        svc = bentoml.Service("iris-classifier", runners=[runner])
        input_spec = Multipart(arr=NumpyNdarray(), annotations=JSON())
        output_spec = Multipart(output=NumpyNdarray(), result=JSON())

        @svc.api(input=input_spec, output=output_spec)
        def predict(arr, annotations):
            res = runner.run(arr)
            return {"output":res, "result":annotations}

    Users then can then serve this service with :code:`bentoml serve`:

    .. code-block:: bash

        % bentoml serve ./sklearn_svc.py:svc --reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "iris-classifier" defined in "sklearn_svc.py"
        [INFO] API Server running on http://0.0.0.0:3000

    Users can then send requests to the newly started services with any client:

    .. tabs::

        .. code-tab:: python

            import requests
            from requests_toolbelt.multipart.encoder import MultipartEncoder

            m = MultipartEncoder(
                fields={'field0': 'value', 'field1': 'value',
                        'field2': ('filename', open('test.json', 'rb'), 'application/json')}
                )

            requests.post('http://0.0.0.0:3000/predict', data=m, headers={'Content-Type': m.content_type})

        .. code-tab:: bash

            % curl -X POST -H "Content-Type: multipart/form-data" -F annotations=@test.json -F arr='[5,4,3,2]' http://0.0.0.0:3000/predict

            --b1d72c201a064ecd92a17a412eb9208e
            Content-Disposition: form-data; name="output"
            content-length: 1
            content-type: application/json

            1
            --b1d72c201a064ecd92a17a412eb9208e
            Content-Disposition: form-data; name="result"
            content-length: 13
            content-type: application/json

            {"foo":"bar"}
            --b1d72c201a064ecd92a17a412eb9208e--

    Args:
        inputs (:code:`Dict[str, IODescriptor]`):
            Dictionary consisting keys as inputs definition for a Multipart
            request/response, values as IODescriptor supported by BentoML. Currently,
            Multipart supports Image, NumpyNdarray, PandasDataFrame, PandasSeries, Text,
            and File.

            Make sure to match the input params in an API function to the keys defined
            under :code:`Multipart`:

            .. code-block:: bash

                +----------------------------------------------------------------+
                |                                                                |
                |   +--------------------------------------------------------+   |
                |   |                                                        |   |
                |   |    Multipart(arr=NumpyNdarray(), annotations=JSON()    |   |
                |   |                                                        |   |
                |   +----------------+-----------------------+---------------+   |
                |                    |                       |                   |
                |                    |                       |                   |
                |                    |                       |                   |
                |                    +----+        +---------+                   |
                |                         |        |                             |
                |         +---------------v--------v---------+                   |
                |         |  def predict(arr, annotations):  |                   |
                |         +----------------------------------+                   |
                |                                                                |
                +----------------------------------------------------------------+

    Returns:
        :obj:`~bentoml._internal.io_descriptors.IODescriptor`: IO Descriptor that Multipart request/response.
    """

    def __init__(self, **inputs: IODescriptor[t.Any]):
        for descriptor in inputs.values():
            if isinstance(descriptor, Multipart):  # pragma: no cover
                raise InvalidArgument(
                    "Multipart IO can not contain nested Multipart IO descriptor"
                )
        self._inputs: dict[str, t.Any] = inputs

    def input_type(
        self,
    ) -> dict[str, t.Type[t.Any] | UnionType | LazyType[t.Any]]:
        res: dict[str, t.Type[t.Any] | UnionType | LazyType[t.Any]] = {}
        for (k, v) in self._inputs.items():
            inp_type = v.input_type()
            if isinstance(inp_type, dict):
                raise TypeError(
                    "A multipart descriptor cannot take a multi-valued I/O descriptor as input"
                )
            res[k] = inp_type

        return res

    def openapi_schema_type(self) -> dict[str, t.Any]:
        return {
            "type": "object",
            "properties": {
                k: io.openapi_schema_type() for (k, io) in self._inputs.items()
            },
        }

    def openapi_request_schema(self) -> dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {"multipart/form-data": {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {"multipart/form-data": {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> dict[str, t.Any]:
        ctype, _ = parse_options_header(request.headers["content-type"])
        if ctype != b"multipart/form-data":
            raise BentoMLException(
                f"{self.__class__.__name__} only accepts `multipart/form-data` as Content-Type header, got {ctype} instead."
            )

        res: dict[str, t.Any] = dict()
        reqs = await populate_multipart_requests(request)

        for k, i in self._inputs.items():
            req = reqs[k]
            v = await i.from_http_request(req)
            res[k] = v
        return res

    async def to_http_response(
        self, obj: dict[str, t.Any], ctx: Context | None = None
    ) -> Response:
        res_mapping: dict[str, Response] = {}
        for k, io_ in self._inputs.items():
            data = obj[k]
            resp = io_.init_http_response()
            res_mapping[k] = await io_.finalize_http_response(resp, data)
        return await concat_to_multipart_response(res_mapping, ctx)



import os
import re
import sys
import inspect
import logging
import importlib
import modulefinder
from typing import List
from typing import Tuple
from unittest.mock import patch

from ..types import PathType
from .pip_pkg import get_all_pip_installed_modules
from ...exceptions import BentoMLException

logger = logging.getLogger(__name__)


def _get_module_src_file(module):
    """
    Return module.__file__, change extension to '.py' if __file__ is ending with '.pyc'
    """
    return module.__file__[:-1] if module.__file__.endswith(".pyc") else module.__file__


def _is_valid_py_identifier(s):
    """
    Return true if string is in a valid python identifier format:

    https://docs.python.org/2/reference/lexical_analysis.html#identifiers
    """
    return re.fullmatch(r"[A-Za-z_][A-Za-z_0-9]*", s) is not None


def _get_module_relative_file_path(module_name, module_file):

    if not os.path.isabs(module_file):
        # For modules within current top level package, module_file here should
        # already be a relative path to the src file
        relative_path = module_file

    elif os.path.split(module_file)[1] == "__init__.py":
        # for module a.b.c in 'some_path/a/b/c/__init__.py', copy file to
        # 'destination/a/b/c/__init__.py'
        relative_path = os.path.join(module_name.replace(".", os.sep), "__init__.py")

    else:
        # for module a.b.c in 'some_path/a/b/c.py', copy file to 'destination/a/b/c.py'
        relative_path = os.path.join(module_name.replace(".", os.sep) + ".py")

    return relative_path


def _get_module(target_module):
    # When target_module is a string, try import it
    if isinstance(target_module, str):
        try:
            target_module = importlib.import_module(target_module)
        except ImportError:
            pass
    return inspect.getmodule(target_module)


def _import_module_from_file(path):
    module_name = path.replace(os.sep, ".")[:-3]
    spec = importlib.util.spec_from_file_location(module_name, path)
    m = importlib.util.module_from_spec(spec)
    return m


# TODO: change this to find_local_py_modules_used(svc: Service)
def find_local_py_modules_used(target_module_file: PathType) -> List[Tuple[str, str]]:
    """Find all local python module dependencies of target_module, and list all the
    local module python files to the destination directory while maintaining the module
    structure unchanged to ensure all imports in target_module still works when loading
    from the destination directory again

    Args:
       path (`Union[str, bytes, os.PathLike]`):
            Path to a python source file

       Returns:
            list of (source file path, target file path) pairs
    """

    target_module = _import_module_from_file(target_module_file)

    try:
        target_module_name = target_module.__spec__.name
    except AttributeError:
        target_module_name = target_module.__name__

    # Find all non pip installed modules must be packaged for target module to run
    exclude_modules = ["bentoml"] + get_all_pip_installed_modules()
    finder = modulefinder.ModuleFinder(excludes=exclude_modules)

    try:
        logger.debug(
            "Searching for local dependant modules of %s:%s",
            target_module_name,
            target_module_file,
        )
        if sys.version_info[0] == 3 and sys.version_info[1] >= 8:
            _find_module = modulefinder._find_module
            _PKG_DIRECTORY = modulefinder._PKG_DIRECTORY

            def _patch_find_module(name, path=None):
                """ref issue: https://bugs.python.org/issue40350"""

                importlib.machinery.PathFinder.invalidate_caches()

                spec = importlib.machinery.PathFinder.find_spec(name, path)

                if spec is not None and spec.loader is None:
                    return None, None, ("", "", _PKG_DIRECTORY)

                return _find_module(name, path)

            with patch.object(modulefinder, "_find_module", _patch_find_module):
                finder.run_script(target_module_file)
        else:
            finder.run_script(target_module_file)
    except SyntaxError:
        # For package with conditional import that may only work with py2
        # or py3, ModuleFinder#run_script will try to compile the source
        # with current python version. And that may result in SyntaxError.
        pass

    if finder.badmodules:
        logger.debug(
            "Find bad module imports that can not be parsed properly: %s",
            finder.badmodules.keys(),
        )

    # Look for dependencies that are not distributed python package, but users'
    # local python code, all other dependencies must be defined with @env
    # decorator when creating a new BentoService class
    user_packages_and_modules = {}
    for name, module in finder.modules.items():
        if hasattr(module, "__file__") and module.__file__ is not None:
            user_packages_and_modules[name] = module

    # Lastly, add target module itself
    user_packages_and_modules[target_module_name] = target_module

    file_list = []
    for module_name, module in user_packages_and_modules.items():
        module_file = _get_module_src_file(module)
        relative_path = _get_module_relative_file_path(module_name, module_file)
        file_list.append((module_file, relative_path))

    return file_list


def copy_local_py_modules(target_module, destination):
    """Find all local python module dependencies of target_module, and copy all the
    local module python files to the destination directory while maintaining the module
    structure unchanged to ensure all imports in target_module still works when loading
    from the destination directory again
    """
    target_module = _get_module(target_module)

    # When target module is defined in interactive session, we can not easily
    # get the class definition into a python module file and distribute it
    if target_module.__name__ == "__main__" and not hasattr(target_module, "__file__"):
        raise BentoMLException(
            "Custom BentoModel class can not be defined in Python interactive REPL, try"
            " writing the class definition to a file and import it."
        )

    try:
        target_module_name = target_module.__spec__.name
    except AttributeError:
        target_module_name = target_module.__name__

    target_module_file = _get_module_src_file(target_module)
    logger.debug(
        "copy_local_py_modules target_module_name: %s, target_module_file: %s",
        target_module_name,
        target_module_file,
    )

    if target_module_name == "__main__":
        # Assuming no relative import in this case
        target_module_file_name = os.path.split(target_module_file)[1]
        target_module_name = target_module_file_name[:-3]  # remove '.py'
        logger.debug(
            "Updating for __main__ module, target_module_name: %s, "
            "target_module_file: %s",
            target_module_name,
            target_module_file,
        )

    # Find all non pip installed modules must be packaged for target module to run
    # exclude_modules = ['bentoml'] + get_all_pip_installed_modules()
    # finder = modulefinder.ModuleFinder(excludes=exclude_modules)
    #
    # try:
    #     logger.debug(
    #         "Searching for local dependant modules of %s:%s",
    #         target_module_name,
    #         target_module_file,
    #     )
    #     if sys.version_info[0] == 3 and sys.version_info[1] >= 8:
    #         _find_module = modulefinder._find_module
    #         _PKG_DIRECTORY = modulefinder._PKG_DIRECTORY
    #
    #         def _patch_find_module(name, path=None):
    #             """ref issue: https://bugs.python.org/issue40350"""
    #
    #             importlib.machinery.PathFinder.invalidate_caches()
    #
    #             spec = importlib.machinery.PathFinder.find_spec(name, path)
    #
    #             if spec is not None and spec.loader is None:
    #                 return None, None, ("", "", _PKG_DIRECTORY)
    #
    #             return _find_module(name, path)
    #
    #         with patch.object(modulefinder, '_find_module', _patch_find_module):
    #             finder.run_script(target_module_file)
    #     else:
    #         finder.run_script(target_module_file)
    # except SyntaxError:
    #     # For package with conditional import that may only work with py2
    #     # or py3, ModuleFinder#run_script will try to compile the source
    #     # with current python version. And that may result in SyntaxError.
    #     pass
    #
    # if finder.badmodules:
    #     logger.debug(
    #         "Find bad module imports that can not be parsed properly: %s",
    #         finder.badmodules.keys(),
    #     )
    #
    # # Look for dependencies that are not distributed python package, but users'
    # # local python code, all other dependencies must be defined with @env
    # # decorator when creating a new BentoService class
    # user_packages_and_modules = {}
    # for name, module in finder.modules.items():
    #     if hasattr(module, "__file__") and module.__file__ is not None:
    #         user_packages_and_modules[name] = module

    # # Remove "__main__" module, if target module is loaded as __main__, it should
    # # be in module_files as (module_name, module_file) in current context
    # if "__main__" in user_packages_and_modules:
    #     del user_packages_and_modules["__main__"]
    #
    # # Lastly, add target module itself
    # user_packages_and_modules[target_module_name] = target_module
    # logger.debug(
    #     "Copying user local python dependencies: %s", user_packages_and_modules
    # )
    #
    # for module_name, module in user_packages_and_modules.items():
    #     module_file = _get_module_src_file(module)
    #     relative_path = _get_module_relative_file_path(module_name, module_file)
    #     target_file = os.path.join(destination, relative_path)
    #
    #     # Create target directory if not exist
    #     Path(os.path.dirname(target_file)).mkdir(parents=True, exist_ok=True)
    #
    #     # Copy module file to BentoArchive for distribution
    #     logger.debug("Copying local python module '%s'", module_file)
    #     copyfile(module_file, target_file)
    #
    # for root, _, files in os.walk(destination):
    #     if "__init__.py" not in files:
    #         logger.debug("Creating empty __init__.py under folder:'%s'", root)
    #         Path(os.path.join(root, "__init__.py")).touch()
    #
    # target_module_relative_path = _get_module_relative_file_path(
    #     target_module_name, target_module_file
    # )
    # logger.debug("Done copying local python dependant modules")
    #
    # return target_module_name, target_module_relative_path



from __future__ import annotations

import os
import typing as t
import logging
from typing import TYPE_CHECKING

import numpy as np

import bentoml
from bentoml import Tag
from bentoml.exceptions import NotFound
from bentoml.exceptions import InvalidArgument
from bentoml.exceptions import MissingDependencyException
from bentoml._internal.models.model import ModelContext

from ..utils.pkg import get_pkg_version

if TYPE_CHECKING:
    from bentoml.types import ModelSignature
    from bentoml.types import ModelSignatureDict

    from .. import external_typing as ext

try:
    import xgboost as xgb
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """xgboost is required in order to use module `bentoml.xgboost`, install
        xgboost with `pip install xgboost`. For more information, refers to
        https://xgboost.readthedocs.io/en/latest/install.html
        """
    )

MODULE_NAME = "bentoml.xgboost"
MODEL_FILENAME = "saved_model.ubj"

logger = logging.getLogger(__name__)


def get(tag_like: str | Tag) -> bentoml.Model:
    """
    Get the BentoML model with the given tag.

    Args:
        tag_like (``str`` ``|`` :obj:`~bentoml.Tag`):
            The tag of the model to retrieve from the model store.
    Returns:
        :obj:`~bentoml.Model`: A BentoML :obj:`~bentoml.Model` with the matching tag.
    Example:

    .. code-block:: python

        import bentoml
        # target model must be from the BentoML model store
        model = bentoml.xgboost.get("my_xgboost_model")
    """
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    return model


def load_model(bento_model: str | Tag | bentoml.Model) -> xgb.core.Booster:
    """
    Load the XGBoost model with the given tag from the local BentoML model store.

    Args:
        bento_model (``str`` ``|`` :obj:`~bentoml.Tag` ``|`` :obj:`~bentoml.Model`):
            Either the tag of the model to get from the store, or a BentoML `~bentoml.Model`
            instance to load the model from.
    Returns:
        :obj:`~xgboost.core.Booster`: The XGBoost model loaded from the model store or BentoML :obj:`~bentoml.Model`.
    Example:

    .. code-block:: python

        import bentoml
        # target model must be from the BentoML model store
        booster = bentoml.xgboost.load_model("my_xgboost_model")
    """  # noqa: LN001
    if not isinstance(bento_model, bentoml.Model):
        bento_model = get(bento_model)
        assert isinstance(bento_model, bentoml.Model)

    model_file = bento_model.path_of(MODEL_FILENAME)
    booster = xgb.core.Booster(model_file=model_file)
    return booster


def save_model(
    name: str,
    model: xgb.core.Booster,
    *,
    signatures: dict[str, ModelSignatureDict] | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> Tag:
    """
    Save an XGBoost model instance to the BentoML model store.

    Args:
        name (``str``):
            The name to give to the model in the BentoML store. This must be a valid
            :obj:`~bentoml.Tag` name.
        model (:obj:`~xgboost.core.Booster`):
            The XGBoost model to be saved.
        signatures (``dict[str, ModelSignatureDict]``, optional):
            Signatures of predict methods to be used. If not provided, the signatures default to
            ``{"predict": {"batchable": False}}``. See :obj:`~bentoml.types.ModelSignature` for more
            details.
        labels (``dict[str, str]``, optional):
            A default set of management labels to be associated with the model. An example is
            ``{"training-set": "data-1"}``.
        custom_objects (``dict[str, Any]``, optional):
            Custom objects to be saved with the model. An example is
            ``{"my-normalizer": normalizer}``.

            Custom objects are currently serialized with cloudpickle, but this implementation is
            subject to change.
        metadata (``dict[str, Any]``, optional):
            Metadata to be associated with the model. An example is ``{"max_depth": 2}``.

            Metadata is intended for display in model management UI and therefore must be a default
            Python type, such as ``str`` or ``int``.
    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the
        user-defined model's name, and a generated `version` by BentoML.

    Example:

    .. code-block:: python

        import xgboost as xgb
        import bentoml

        # read in data
        dtrain = xgb.DMatrix('demo/data/agaricus.txt.train')
        dtest = xgb.DMatrix('demo/data/agaricus.txt.test')
        # specify parameters via map
        param = dict(max_depth=2, eta=1, objective='binary:logistic')
        num_round = 2
        bst = xgb.train(param, dtrain, num_round)
        ...

        # `save` the booster to BentoML modelstore:
        tag = bentoml.xgboost.save_model("my_xgboost_model", bst, booster_params=param)
    """  # noqa: LN001
    context: ModelContext = ModelContext(
        framework_name="xgboost",
        framework_versions={"xgboost": get_pkg_version("xgboost")},
    )

    if signatures is None:
        logger.info(
            'Using default model signature `{"predict": {"batchable": False}}` for XGBoost model'
        )
        signatures = {
            "predict": {"batchable": False},
        }

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        signatures=signatures,
        labels=labels,
        custom_objects=custom_objects,
        metadata=metadata,
        context=context,
    ) as bento_model:
        model.save_model(bento_model.path_of(MODEL_FILENAME))  # type: ignore (incomplete XGBoost types)

        return bento_model.tag


def get_runnable(bento_model: bentoml.Model) -> t.Type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class XGBoostRunnable(bentoml.Runnable):
        SUPPORT_NVIDIA_GPU = True
        SUPPORT_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()
            self.model = load_model(bento_model)

            # check for resources
            available_gpus = os.getenv("NVIDIA_VISIBLE_DEVICES")
            if available_gpus is not None and available_gpus != "":
                self.model.set_param({"predictor": "gpu_predictor", "gpu_id": 0})  # type: ignore (incomplete XGBoost types)
            else:
                nthreads = os.getenv("OMP_NUM_THREADS")
                if nthreads is not None and nthreads != "":
                    nthreads = max(int(nthreads), 1)
                else:
                    nthreads = 1
                self.model.set_param({"predictor": "cpu_predictor", "nthread": nthreads})  # type: ignore (incomplete XGBoost types)

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                try:
                    self.predict_fns[method_name] = getattr(self.model, method_name)
                except AttributeError:
                    raise InvalidArgument(
                        f"No method with name {method_name} found for XGBoost model of type {self.model.__class__}"
                    )

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(
            self: XGBoostRunnable,
            input_data: ext.NpNDArray
            | ext.PdDataFrame,  # TODO: add support for DMatrix
        ) -> ext.NpNDArray:
            dmatrix = xgb.DMatrix(input_data)

            res = self.predict_fns[method_name](dmatrix)
            return np.asarray(res)  # type: ignore (incomplete np types)

        XGBoostRunnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return XGBoostRunnable



from .schemas import CliEvent
from .schemas import ModelSaveEvent
from .schemas import BentoBuildEvent
from .schemas import ServeUpdateEvent
from .cli_events import cli_events_map
from .usage_stats import track
from .usage_stats import ServeInfo
from .usage_stats import track_serve
from .usage_stats import get_serve_info
from .usage_stats import BENTOML_DO_NOT_TRACK

__all__ = [
    "track",
    "track_serve",
    "get_serve_info",
    "ServeInfo",
    "BENTOML_DO_NOT_TRACK",
    "CliEvent",
    "ModelSaveEvent",
    "BentoBuildEvent",
    "ServeUpdateEvent",
    "cli_events_map",
]



import os
import ast
import sys
import typing as t
import logging
import pkgutil
import zipfile
import zipimport
from typing import TYPE_CHECKING
from collections import defaultdict

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata
from packaging.requirements import Requirement

if TYPE_CHECKING:
    from ..service import Service

EPP_NO_ERROR = 0
EPP_PKG_NOT_EXIST = 1
EPP_PKG_VERSION_MISMATCH = 2

__mm = None

logger = logging.getLogger(__name__)


def split_requirement(requirement: str) -> t.Tuple[str, str]:
    """
    Split requirements. 'bentoml>=1.0.0' -> ['bentoml', '>=1.0.0']
    """
    req = Requirement(requirement)
    name = req.name.replace("-", "_")
    return name, str(req.specifier)


def packages_distributions() -> t.Dict[str, t.List[str]]:
    """Return a mapping of top-level packages to their distributions. We're
    inlining this helper from the importlib_metadata "backport" here, since
    it's not available in the builtin importlib.metadata.
    """
    pkg_to_dist = defaultdict(list)
    for dist in importlib_metadata.distributions():
        for pkg in (dist.read_text("top_level.txt") or "").split():
            pkg_to_dist[pkg].append(dist.metadata["Name"])
    return dict(pkg_to_dist)


def parse_requirement_string(rs):
    name, _, version = rs.partition("==")
    return name, version


def verify_pkg(pkg_req):
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.verify_pkg(pkg_req)


def seek_pip_packages(target_py_file_path):
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.seek_pip_packages(target_py_file_path)


def get_pkg_version(pkg_name):
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.pip_pkg_map.get(pkg_name, None)


def get_zipmodules():
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.zip_modules


def get_all_pip_installed_modules():
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()

    installed_modules = list(
        # local modules are the ones imported from current directory, either from a
        # module.py file or a module directory that contains a `__init__.py` file
        filter(lambda m: not m.is_local, __mm.searched_modules.values())
    )
    return list(map(lambda m: m.name, installed_modules))


class ModuleInfo(object):
    def __init__(self, name, path, is_local, is_pkg):
        super(ModuleInfo, self).__init__()
        self.name = name
        self.path = path
        self.is_local = is_local
        self.is_pkg = is_pkg


class ModuleManager(object):
    def __init__(self):
        super(ModuleManager, self).__init__()
        self.pip_pkg_map = {}
        self.pip_module_map = {}
        self.setuptools_module_set = set()
        self.nonlocal_package_path = set()

        import pkg_resources

        for dist in pkg_resources.working_set:  # pylint: disable=not-an-iterable
            module_path = dist.module_path or dist.location
            if not module_path:
                # Skip if no module path was found for pkg distribution
                continue

            if os.path.realpath(module_path) != os.getcwd():
                # add to nonlocal_package path only if it's not current directory
                self.nonlocal_package_path.add(module_path)

            self.pip_pkg_map[dist._key] = dist._version
            for mn in dist._get_metadata("top_level.txt"):
                if dist._key != "setuptools":
                    self.pip_module_map.setdefault(mn, []).append(
                        (dist._key, dist._version)
                    )
                else:
                    self.setuptools_module_set.add(mn)

        self.searched_modules = {}
        self.zip_modules: t.Dict[str, zipimport.zipimporter] = {}
        for m in pkgutil.iter_modules():
            if m.name not in self.searched_modules:
                if isinstance(m.module_finder, zipimport.zipimporter):
                    logger.info(f"Detected zipimporter {m.module_finder}")
                    path = m.module_finder.archive
                    self.zip_modules[path] = m.module_finder
                else:
                    path = m.module_finder.path
                is_local = self.is_local_path(path)
                self.searched_modules[m.name] = ModuleInfo(
                    m.name, path, is_local, m.ispkg
                )

    def verify_pkg(self, pkg_req):
        if pkg_req.name not in self.pip_pkg_map:
            # package does not exist in the current python session
            return EPP_PKG_NOT_EXIST

        if self.pip_pkg_map[pkg_req.name] not in pkg_req.specifier:
            # package version being used in the current python session does not meet
            # the specified package version requirement
            return EPP_PKG_VERSION_MISMATCH

        return EPP_NO_ERROR

    def seek_pip_packages(self, target_py_file_path):
        logger.debug("target py file path: %s", target_py_file_path)
        work = DepSeekWork(self, target_py_file_path)
        work.do()
        requirements = {}
        for _, pkg_info_list in work.dependencies.items():
            for pkg_name, pkg_version in pkg_info_list:
                requirements[pkg_name] = pkg_version

        return requirements, work.unknown_module_set

    def is_local_path(self, path):
        if path in self.nonlocal_package_path:
            return False

        dir_name = os.path.split(path)[1]

        if (
            "site-packages" in path
            or "anaconda" in path
            or path.endswith("packages")
            or dir_name == "bin"
            or dir_name.startswith("lib")
            or dir_name.startswith("python")
            or dir_name.startswith("plat")
        ):
            self.nonlocal_package_path.add(path)
            return False

        return True


class DepSeekWork(object):
    def __init__(self, module_manager, target_py_file_path):
        super(DepSeekWork, self).__init__()
        self.module_manager = module_manager
        self.target_py_file_path = target_py_file_path

        self.dependencies = {}
        self.unknown_module_set = set()
        self.parsed_module_set = set()

    def do(self):
        self.seek_in_file(self.target_py_file_path)

    def seek_in_file(self, file_path):
        try:
            with open(file_path) as f:
                content = f.read()
        except UnicodeDecodeError:
            with open(file_path, encoding="utf-8") as f:
                content = f.read()
        self.seek_in_source(content)

    def seek_in_source(self, content):
        # Extract all dependency modules by searching through the trees of the Python
        # abstract syntax grammar with Python's built-in ast module
        tree = ast.parse(content)
        import_set = set()
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for name in node.names:
                    import_set.add(name.name.partition(".")[0])
            elif isinstance(node, ast.ImportFrom):
                if node.module is not None and node.level == 0:
                    import_set.add(node.module.partition(".")[0])
        logger.debug("import set: %s", import_set)
        for module_name in import_set:
            # Avoid parsing BentoML when BentoML is imported from local source code repo
            if module_name == "bentoml":
                continue
            if module_name in self.parsed_module_set:
                continue
            self.parsed_module_set.add(module_name)

            if module_name in self.module_manager.searched_modules:
                m = self.module_manager.searched_modules[module_name]
                if m.is_local:
                    # Recursively search dependencies in sub-modules
                    if m.path in self.module_manager.zip_modules:
                        self.seek_in_zip(m.path)
                    elif m.is_pkg:
                        self.seek_in_dir(os.path.join(m.path, m.name))
                    else:
                        self.seek_in_file(os.path.join(m.path, "{}.py".format(m.name)))
                else:
                    # check if the package has already been added to the list
                    if (
                        module_name in self.module_manager.pip_module_map
                        and module_name not in self.dependencies
                        and module_name not in self.module_manager.setuptools_module_set
                    ):
                        self.dependencies[
                            module_name
                        ] = self.module_manager.pip_module_map[module_name]
            else:
                if module_name in self.module_manager.pip_module_map:
                    if module_name not in self.dependencies:
                        # In some special cases, the pip-installed module can not
                        # be located in the searched_modules
                        self.dependencies[
                            module_name
                        ] = self.module_manager.pip_module_map[module_name]
                else:
                    if module_name not in sys.builtin_module_names:
                        self.unknown_module_set.add(module_name)

    def seek_in_dir(self, dir_path):
        for path, dir_list, file_list in os.walk(dir_path):
            for file_name in file_list:
                if not file_name.endswith(".py"):
                    continue
                self.seek_in_file(os.path.join(path, file_name))
            for dir_name in dir_list:
                if dir_name in ["__pycache__", ".ipynb_checkpoints"]:
                    continue
                self.seek_in_dir(os.path.join(path, dir_name))

    def seek_in_zip(self, zip_path):
        with zipfile.ZipFile(zip_path) as zf:
            for module_path in zf.infolist():
                filename = module_path.filename
                if filename.endswith(".py"):
                    logger.debug("Seeking modules in zip %s", filename)
                    content = self.module_manager.zip_modules[zip_path].get_source(
                        filename.replace(".py", "")
                    )
                    self.seek_in_source(content)


def lock_pypi_versions(package_list: t.List[str]) -> t.List[str]:
    """
    Lock versions of pypi packages in current virtualenv

    Args:
        package_list List[str]:
            List contains package names

    Raises:
        ValueError: if one package in `package_list` is not
         available in current virtualenv

    Returns:

        list of lines for requirements.txt

        Example Results:

        * ['numpy==1.20.3', 'pandas==1.2.4', 'scipy==1.4.1']
    """
    pkgs_with_version = []

    for pkg in package_list:
        version = get_pkg_version(pkg)
        print(pkg, version)
        if version:
            pkg_line = f"{pkg}=={version}"
            pkgs_with_version.append(pkg_line)
        else:
            # better warning or throw an exception?
            raise ValueError(f"package {pkg} is not available in current virtualenv")

    return pkgs_with_version


def with_pip_install_options(
    package_lines: t.List[str],
    index_url: t.Optional[str] = None,
    extra_index_url: t.Optional[str] = None,
    find_links: t.Optional[str] = None,
) -> t.List[str]:
    """
    Lock versions of pypi packages in current virtualenv

    Args:
        package_lines List[str]:
            List contains items each representing one line of requirements.txt

        index_url Optional[str]:
            value of --index-url

        extra_index_url Optional[str]:
            value of --extra_index-url

        find_links Optional[str]:
            value of --find-links

    Returns:

        list of lines for requirements.txt

        Example Results:

        * ['pandas==1.2.4 --index-url=https://mirror.baidu.com/pypi/simple',
            'numpy==1.20.3 --index-url=https://mirror.baidu.com/pypi/simple']
    """

    options = []
    if index_url:
        options.append(f"--index-url={index_url}")
    if extra_index_url:
        options.append(f"--extra-index-url={extra_index_url}")
    if find_links:
        options.append(f"--find-links={find_links}")

    if not options:
        return package_lines

    option_str = " ".join(options)
    pkgs_with_options = [pkg + " " + option_str for pkg in package_lines]
    return pkgs_with_options


def find_required_pypi_packages(
    svc: "Service", lock_versions: bool = True
) -> t.List[str]:
    """
    Find required pypi packages in a python source file

    Args:
        path (`Union[str, bytes, os.PathLike]`):
            Path to a python source file

        lock_versions bool:
            if the versions of packages should be locked

    Returns:

        list of lines for requirements.txt

        Example Results:

        * ['numpy==1.20.3', 'pandas==1.2.4']
        * ['numpy', 'pandas']
    """
    module_name = svc.__module__
    module = sys.modules[module_name]
    reqs, unknown_modules = seek_pip_packages(module.__file__)
    for module_name in unknown_modules:
        logger.warning("unknown package dependency for module: %s", module_name)

    if lock_versions:
        pkg_lines = ["%s==%s" % pkg for pkg in reqs.items()]
    else:
        pkg_lines = list(reqs.keys())

    return pkg_lines



from __future__ import annotations

import typing as t
import logging
import pathlib
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from circus.plugins import CircusPlugin  # type: ignore[reportMissingTypeStubs]

if TYPE_CHECKING:
    from circus.arbiter import Arbiter  # type: ignore[reportMissingTypeStubs]
    from circus.sockets import CircusSocket  # type: ignore[reportMissingTypeStubs]
    from circus.watcher import Watcher  # type: ignore[reportMissingTypeStubs]

logger = logging.getLogger(__name__)


def create_circus_socket_from_uri(
    uri: str,
    *args: t.Any,
    name: str = "",
    **kwargs: t.Any,
) -> CircusSocket:
    from circus.sockets import CircusSocket  # type: ignore[reportMissingTypeStubs]

    from bentoml._internal.utils.uri import uri_to_path

    parsed = urlparse(uri)
    if parsed.scheme in ("file", "unix"):
        return CircusSocket(
            name=name,
            path=uri_to_path(uri),
            *args,
            **kwargs,
        )
    elif parsed.scheme == "tcp":
        return CircusSocket(
            name=name,
            host=parsed.hostname,
            port=parsed.port,
            *args,
            **kwargs,
        )
    else:
        raise ValueError(f"Unsupported URI scheme: {parsed.scheme}")


def create_standalone_arbiter(watchers: list[Watcher], **kwargs: t.Any) -> Arbiter:
    from circus.arbiter import Arbiter  # type: ignore[reportMissingTypeStubs]

    from . import reserve_free_port

    with reserve_free_port() as endpoint_port:
        with reserve_free_port() as pubsub_port:
            return Arbiter(
                watchers,
                endpoint=f"tcp://127.0.0.1:{endpoint_port}",
                pubsub_endpoint=f"tcp://127.0.0.1:{pubsub_port}",
                **kwargs,
            )


# TODO: use svc.build_args.include/exclude as default files to watch
# TODO: watch changes in model store when "latest" model tag is used


class BentoChangeReloader(CircusPlugin):
    """
    A circus plugin that reloads the BentoService when the service code changes.


    Args:
        working_dir: The directory of the bento service.
        reload_delay: The delay in seconds between checking for changes.
    """

    name = "bento_change_reloader"
    config: dict[str, t.Any]

    def __init__(self, *args: t.Any, **config: t.Any):
        assert "bento_identifier" in config, "`bento_identifier` is required"
        assert "working_dir" in config, "`working_dir` is required"

        super().__init__(*args, **config)  # type: ignore (unfinished types for circus)

        self.name = self.config.get("name")
        working_dir: str = self.config["working_dir"]

        # circus/plugins/__init__.py:282 -> converts all given configs to dict[str, str]
        self.reload_delay: float = float(self.config.get("reload_delay", 1))
        self.file_watcher = PyFileChangeWatcher([working_dir])

    def look_after(self):
        if self.file_watcher.is_file_changed():
            logger.info("Restarting...")
            self.call("restart", name="*")  # type: ignore
            self.file_watcher.reset()

    def handle_init(self):
        from tornado import ioloop

        self.period = ioloop.PeriodicCallback(self.look_after, self.reload_delay * 1000)
        self.period.start()

    def handle_stop(self):
        self.period.stop()

    def handle_recv(self, data: t.Any):
        pass


class PyFileChangeWatcher:
    def __init__(
        self,
        watch_dirs: list[pathlib.Path] | list[str] | None = None,
    ) -> None:
        self.mtimes: dict[pathlib.Path, float] = {}
        if not watch_dirs:
            watch_dirs = [pathlib.Path.cwd()]
        self.watch_dirs = [
            pathlib.Path(d) if isinstance(d, str) else d for d in watch_dirs
        ]
        logger.info(f"Watching directories: {', '.join(map(str, self.watch_dirs))}")

    def is_file_changed(self) -> bool:
        for file in self.iter_files():
            try:
                mtime = file.stat().st_mtime
            except OSError:  # pragma: nocover
                continue

            old_time = self.mtimes.get(file)
            if old_time is None:
                self.mtimes[file] = mtime
                continue
            elif mtime > old_time:
                display_path = str(file)
                try:
                    display_path = str(file.relative_to(pathlib.Path.cwd()))
                except ValueError:
                    pass
                message = "Detected file change in '%s'"
                logger.warning(message, display_path)
                return True
        return False

    def reset(self) -> None:
        self.mtimes = {}

    def iter_files(self) -> t.Iterator[pathlib.Path]:
        for reload_dir in self.watch_dirs:
            for path in list(reload_dir.rglob("*.py")):
                yield path.resolve()



import logging
from typing import Optional
from urllib.parse import urljoin

import requests

from .schemas import UserSchema
from .schemas import BentoSchema
from .schemas import ModelSchema
from .schemas import schema_to_json
from .schemas import schema_from_json
from .schemas import CreateBentoSchema
from .schemas import CreateModelSchema
from .schemas import UpdateBentoSchema
from .schemas import OrganizationSchema
from .schemas import BentoRepositorySchema
from .schemas import ModelRepositorySchema
from .schemas import FinishUploadBentoSchema
from .schemas import FinishUploadModelSchema
from .schemas import CreateBentoRepositorySchema
from .schemas import CreateModelRepositorySchema
from ...exceptions import YataiRESTApiClientError

logger = logging.getLogger(__name__)


class YataiRESTApiClient:
    def __init__(self, endpoint: str, api_token: str) -> None:
        self.endpoint = endpoint
        self.session = requests.Session()
        self.session.headers.update(
            {
                "X-YATAI-API-TOKEN": api_token,
                "Content-Type": "application/json",
            }
        )

    def _is_not_found(self, resp: requests.Response) -> bool:
        # Forgive me, I don't know how to map the error returned by gorm to juju/errors
        return resp.status_code == 400 and "record not found" in resp.text

    def _check_resp(self, resp: requests.Response) -> None:
        if resp.status_code != 200:
            raise YataiRESTApiClientError(
                f"request failed with status code {resp.status_code}: {resp.text}"
            )

    def get_current_user(self) -> Optional[UserSchema]:
        url = urljoin(self.endpoint, "/api/v1/auth/current")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, UserSchema)

    def get_current_organization(self) -> Optional[OrganizationSchema]:
        url = urljoin(self.endpoint, "/api/v1/current_org")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, OrganizationSchema)

    def get_bento_repository(
        self, bento_repository_name: str
    ) -> Optional[BentoRepositorySchema]:
        url = urljoin(
            self.endpoint, f"/api/v1/bento_repositories/{bento_repository_name}"
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoRepositorySchema)

    def create_bento_repository(
        self, req: CreateBentoRepositorySchema
    ) -> BentoRepositorySchema:
        url = urljoin(self.endpoint, "/api/v1/bento_repositories")
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoRepositorySchema)

    def get_bento(
        self, bento_repository_name: str, version: str
    ) -> Optional[BentoSchema]:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}",
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def create_bento(
        self, bento_repository_name: str, req: CreateBentoSchema
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint, f"/api/v1/bento_repositories/{bento_repository_name}/bentos"
        )
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def update_bento(
        self, bento_repository_name: str, version: str, req: UpdateBentoSchema
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def presign_bento_upload_url(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_upload_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def presign_bento_download_url(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_download_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def start_upload_bento(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/start_upload",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def finish_upload_bento(
        self, bento_repository_name: str, version: str, req: FinishUploadBentoSchema
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/finish_upload",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def get_model_repository(
        self, model_repository_name: str
    ) -> Optional[ModelRepositorySchema]:
        url = urljoin(
            self.endpoint, f"/api/v1/model_repositories/{model_repository_name}"
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelRepositorySchema)

    def create_model_repository(
        self, req: CreateModelRepositorySchema
    ) -> ModelRepositorySchema:
        url = urljoin(self.endpoint, "/api/v1/model_repositories")
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelRepositorySchema)

    def get_model(
        self, model_repository_name: str, version: str
    ) -> Optional[ModelSchema]:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}",
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def create_model(
        self, model_repository_name: str, req: CreateModelSchema
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint, f"/api/v1/model_repositories/{model_repository_name}/models"
        )
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def presign_model_upload_url(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_upload_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def presign_model_download_url(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_download_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def start_upload_model(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/start_upload",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def finish_upload_model(
        self, model_repository_name: str, version: str, req: FinishUploadModelSchema
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/finish_upload",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)



from __future__ import annotations

import typing as t
import functools
from typing import TYPE_CHECKING

from bentoml._internal.runner.utils import Params
from bentoml._internal.runner.container import Payload
from bentoml._internal.runner.container import AutoContainer

from . import RunnerHandle

if TYPE_CHECKING:
    from ..runner import Runner
    from ..runner import RunnerMethod
    from ..runnable import Runnable

    P = t.ParamSpec("P")
    R = t.TypeVar("R")


class LocalRunnerRef(RunnerHandle):
    _runnable: Runnable

    def __init__(self, runner: Runner) -> None:  # pylint: disable=super-init-not-called
        self._runnable = runner.runnable_class()

    def run_method(
        self,
        __bentoml_method: RunnerMethod[t.Any, P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        if __bentoml_method.config.batchable:
            inp_batch_dim = __bentoml_method.config.batch_dim[0]

            payload_params = Params[Payload](*args, **kwargs).map(
                lambda arg: AutoContainer.to_payload(arg, batch_dim=inp_batch_dim)
            )

            params_iter = iter(payload_params.items())
            size = next(params_iter)[1].batch_size
            for param in params_iter:
                if param[1].batch_size != size:
                    raise ValueError(
                        "All batchable arguments must have the same batch size."
                    )
        return getattr(self._runnable, __bentoml_method.name)(*args, **kwargs)

    async def async_run_method(
        self,
        __bentoml_method: RunnerMethod[t.Any, P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        import anyio

        method = getattr(self._runnable, __bentoml_method.name)
        return await anyio.to_thread.run_sync(
            functools.partial(method, **kwargs), *args, limiter=anyio.CapacityLimiter(1)
        )



from __future__ import annotations

import typing as t
import contextvars
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

import attr
import starlette.datastructures

from .utils.http import Cookie

if TYPE_CHECKING:
    import starlette.requests
    import starlette.responses


class Metadata(t.Mapping[str, str], ABC):
    @abstractmethod
    def __setitem__(self, key: str, value: str) -> None:
        """
        Set the header ``key`` to ``value``, removing any duplicate entries.
        Retains insertion order.
        """

    @abstractmethod
    def __delitem__(self, key: str) -> None:
        """
        Remove the header ``key``.
        """

    @abstractmethod
    def __ior__(self, other: t.Mapping[t.Any, t.Any]) -> Metadata:
        """
        Updates this metadata with the contents of ``other``
        """

    @abstractmethod
    def __or__(self, other: t.Mapping[t.Any, t.Any]) -> Metadata:
        """
        Returns a new metadata object with the contents of this metadata object updated with the contents of ``other``
        """

    @abstractmethod
    def update(self, other: t.Mapping[t.Any, t.Any]) -> None:
        """
        Sets all the headers in ``other`` in this object.

        For example, if this object is ``{"my-header": "my-value", "my-other-header": "my-other-value"}``
        and other is {"my-header": 3¸ "other-header": 4}
        """

    @abstractmethod
    def setdefault(self, key: str, value: str) -> str:
        """
        If the header ``key`` does not exist, then set it to ``value``.
        Returns the header value.
        """

    @abstractmethod
    def append(self, key: str, value: str) -> None:
        """
        Append a header, preserving any duplicate entries.
        """

    @abstractmethod
    def mutablecopy(self) -> Metadata:
        """
        Returns a copy of this metadata object.
        """


@attr.define
class InferenceApiContext:
    request: "RequestContext"
    response: "ResponseContext"

    def __init__(self, request: "RequestContext", response: "ResponseContext"):
        self.request = request
        self.response = response

    @staticmethod
    def from_http(request: "starlette.requests.Request") -> "InferenceApiContext":
        request_ctx = InferenceApiContext.RequestContext.from_http(request)
        response_ctx = InferenceApiContext.ResponseContext()

        return InferenceApiContext(request_ctx, response_ctx)

    @attr.define
    class RequestContext:
        metadata: Metadata
        headers: Metadata

        def __init__(self, metadata: Metadata):
            self.metadata = metadata
            self.headers = metadata

        @staticmethod
        def from_http(
            request: "starlette.requests.Request",
        ) -> "InferenceApiContext.RequestContext":
            return InferenceApiContext.RequestContext(request.headers)  # type: ignore (coercing Starlette headers to Metadata)

    @attr.define
    class ResponseContext:
        metadata: Metadata
        cookies: list[Cookie]
        headers: Metadata
        status_code: int = 200

        def __init__(self):
            self.metadata = starlette.datastructures.MutableHeaders()  # type: ignore (coercing Starlette headers to Metadata)
            self.headers = self.metadata  # type: ignore (coercing Starlette headers to Metadata)

        def set_cookie(
            self,
            key: str,
            value: str,
            max_age: int | None = None,
            expires: int | None = None,
            path: str = "/",
            domain: str | None = None,
            secure: bool = False,
            httponly: bool = False,
            samesite: str = "lax",
        ):
            self.cookies.append(
                Cookie(
                    key,
                    value,
                    max_age,
                    expires,
                    path,
                    domain,
                    secure,
                    httponly,
                    samesite,
                )
            )


class ServiceTraceContext:
    def __init__(self) -> None:
        self._request_id_var = contextvars.ContextVar(
            "_request_id_var", default=t.cast("t.Optional[int]", None)
        )

    @property
    def trace_id(self) -> t.Optional[int]:
        from opentelemetry import trace

        span = trace.get_current_span()
        if span is None:
            return None
        return span.get_span_context().trace_id

    @property
    def span_id(self) -> t.Optional[int]:
        from opentelemetry import trace  # type: ignore

        span = trace.get_current_span()
        if span is None:
            return None
        return span.get_span_context().span_id

    @property
    def request_id(self) -> t.Optional[int]:
        """
        Different from span_id, request_id is unique for each inbound request.
        """
        return self._request_id_var.get()

    @request_id.setter
    def request_id(self, request_id: t.Optional[int]) -> None:
        self._request_id_var.set(request_id)

    @request_id.deleter
    def request_id(self) -> None:
        self._request_id_var.set(None)


trace_context = ServiceTraceContext()



import os
import typing as t
import logging
import importlib
from typing import TYPE_CHECKING
from pathlib import Path
from distutils.dir_util import copy_tree

import yaml
from simple_di import inject
from simple_di import Provide

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..utils import LazyLoader
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..bento.pip_pkg import split_requirement
from ..bento.pip_pkg import packages_distributions
from .common.model_runner import BaseModelRunner
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:  # pragma: no cover
    from spacy.vocab import Vocab
    from thinc.config import Config
    from spacy.tokens.doc import Doc

    from ..models import ModelStore

try:
    import spacy
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """\
        `spacy` is required to use with `bentoml.spacy`.
        Instruction: Refers to https://spacy.io/usage for more information.
        """
    )

MODULE_NAME = "bentoml.spacy"


torch = LazyLoader("torch", globals(), "torch")
tensorflow = LazyLoader("tensorflow", globals(), "tensorflow")

_TORCH_TF_WARNING = """\
It is recommended that if you want to run SpaCy with {framework}
 model that you also have `{package}` installed.
Refers to {link} for more information.

We also detected that you choose to run on GPUs, thus in order to utilize
 BentoML Runners features with GPUs you should also install `{package}` with
 CUDA support.
"""

PROJECTS_CMD_NOT_SUPPORTED = [
    "assets",
    "document",
    "dvc",
    "push",
    "run",
]

logger = logging.getLogger(__name__)


@inject
def load_project(
    tag: Tag,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> str:
    model = model_store.get(tag)
    if model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    if "projects_uri" in model.info.options:
        logger.warning(
            "We will only returns the path of projects saved under BentoML modelstore"
            " and leave projects interaction for users to decide what they want to do."
            " Refers to https://spacy.io/api/cli#project for more information."
        )
        return os.path.join(model.path, model.info.options["target_path"])
    raise BentoMLException(
        "Cannot use `bentoml.spacy.load_project()` to load non Spacy Projects. If your"
        " model is not a Spacy projects use `bentoml.spacy.load()` instead."
    )


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    vocab: t.Union["Vocab", bool] = True,  # type: ignore[reportUnknownParameterType]
    disable: t.Sequence[str] = tuple(),
    exclude: t.Sequence[str] = tuple(),
    config: t.Union[t.Dict[str, t.Any], "Config", None] = None,
) -> "spacy.language.Language":
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.
        vocab (:code:`Union[spacy.vocab.Vocab, bool]`, `optional`, defaults to `True`):
            Optional vocab to pass in on initialization. If True, a new Vocab object will be created.
        disable (`Sequence[str]`, `optional`):
            Names of pipeline components to disable.
        exclude (`Sequence[str]`, `optional`):
            Names of pipeline components to exclude. Excluded
            components won't be loaded.
        config (:code:`Union[Dict[str, Any], spacy.Config]`, `optional`):
            Config overrides as nested dict or dict
            keyed by section values in dot notation.

    Returns:
        :obj:`spacy.language.Language`: an instance of :obj:`spacy.Language` from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml

        model = bentoml.spacy.load('custom_roberta')
    """
    model = model_store.get(tag)
    if model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )

    if "projects_uri" in model.info.options:
        raise BentoMLException(
            "Cannot use `bentoml.spacy.load()` to load Spacy Projects. Use"
            " `bentoml.spacy.load_project()` instead."
        )
    required = model.info.options["pip_package"]

    try:
        _ = importlib.import_module(required)
    except ModuleNotFoundError:
        try:
            from spacy.cli.download import download

            # TODO: move this to runner on startup hook
            download(required)
        except (SystemExit, Exception):  # pylint: disable=broad-except
            logger.warning(
                f"{required} cannot be downloaded as pip package. If this"
                " is a custom pipeline there is nothing to worry about."
                " If this is a pretrained model provided by Explosion make"
                " sure that you save the correct package and model to BentoML"
                " via `bentoml.spacy.save()`"
            )
    try:
        # check if pipeline has additional requirements then all related
        # pip package has been installed correctly.
        additional = model.info.options["additional_requirements"]
        not_existed = list()  # type: t.List[str]
        dists = packages_distributions()
        for module_name in additional:
            mod, _ = split_requirement(module_name)
            if mod not in dists:
                not_existed.append(module_name)
            if len(not_existed) > 0:
                raise MissingDependencyException(
                    f"`{','.join(not_existed)}` is required by `{tag}`."
                )
    except KeyError:
        pass
    import spacy.util

    return spacy.util.load_model(
        model.path,
        vocab=vocab,
        disable=disable,
        exclude=exclude,
        config=config if config else {},
    )


def save(
    name: str,
    model: "spacy.language.Language",
    *,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Union[None, t.Dict[str, t.Any]] = None,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`spacy.language.Language`):
            Instance of model to be saved
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import spacy
        import bentoml.spacy
        nlp = spacy.load("en_core_web_trf")

        # custom training or layers here

        bentoml.spacy.save("spacy_roberta", nlp)
    """  # noqa
    context: t.Dict[str, t.Any] = {
        "framework_name": "spacy",
        "pip_dependencies": [f"spacy=={get_pkg_version('spacy')}"],
    }

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        options=None,
        context=context,
        labels=labels,
        custom_objects=custom_objects,
        metadata=metadata,
    ) as _model:

        meta = model.meta
        pip_package = f"{meta['lang']}_{meta['name']}"
        _model.info.options = {"pip_package": pip_package}
        if "requirements" in meta:
            _model.info.options["additional_requirements"] = meta["requirements"]
        model.to_disk(_model.path)

        return _model.tag


@inject
def projects(
    save_name: str,
    tasks: str,
    name: t.Optional[str] = None,
    repo_or_store: t.Optional[str] = None,
    remotes_config: t.Optional[t.Dict[str, t.Dict[str, str]]] = None,
    *,
    branch: t.Optional[str] = None,
    sparse_checkout: bool = False,
    verbose: bool = True,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Enables users to use :code:`spacy cli` and integrate SpaCy `Projects <https://spacy.io/usage/projects>`_ to BentoML.

    Args:
        save_name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        tasks (:code:`str`):
            Given SpaCy CLI tasks. Currently only support :code:`pull` and :code:`clone`
        repo_or_store(:code:`str`, `optional`, defaults to `None`):
            URL of Git repo or given S3 store containing project templates.
        model (`spacy.language.Language`):
            Instance of model to be saved
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.
        branch (:code:`str`, `optional`, defaults to `None`):
            The branch to clone from. If not specified, defaults to :code:`main` branch
        verbose (`bool`, `optional`, default to :code:`True`):
            Verbosely post all logs.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    .. warning::

       This is an **EXPERIMENTAL** API as it is subjected to change. We are also looking for feedback.

    Examples:

    .. code-block:: python

        import bentoml

        clone_tag = bentoml.spacy.projects(
            "test_spacy_project",
            "clone",
            name="integrations/huggingface_hub",
            repo_or_store="https://github.com/aarnphm/bentoml-spacy-projects-integration-tests",
        )
        project_path = bentoml.spacy.load_project(clone_tag)


        project_yml = {
            "remotes": {
                "default": "https://github.com/aarnphm/bentoml-spacy-projects-integration-tests/tree/v3/pipelines/tagger_parser_ud",
            }
        }
        pull_tag = bentoml.spacy.projects("test_pull", "pull", remotes_config=project_yml)
        project_path = bentoml.spacy.load_project(pull_tag)
    """
    # EXPERIMENTAL: note that these functions are direct modified implementation
    # from spacy internal API. Subject to change, use with care!
    from spacy.cli.project.pull import project_pull
    from spacy.cli.project.clone import project_clone

    if repo_or_store is None:
        repo_or_store = getattr(spacy.about, "__projects__")
    if branch is None:
        branch = getattr(spacy.about, "__projects_branch__")

    if tasks in PROJECTS_CMD_NOT_SUPPORTED:
        raise BentoMLException(
            """\
        BentoML only supports `clone` and `pull` for
         git and remote storage from SpaCy CLI to
         save into modelstore. Refers to
         https://spacy.io/api/cli#project for more
         information on SpaCy Projects.
        """
        )
    context: t.Dict[str, t.Any] = {
        "framework_name": "spacy",
        "pip_dependencies": [f"spacy=={get_pkg_version('spacy')}"],
        "tasks": tasks,
    }
    with bentoml.models.create(
        save_name,
        module=MODULE_NAME,
        options=None,
        context=context,
        metadata=metadata,
    ) as _model:
        output_path = _model.path_of(SAVE_NAMESPACE)
        _model.info.options = {
            "projects_uri": repo_or_store,
            "target_path": SAVE_NAMESPACE,
        }
        if tasks == "clone":
            # TODO: update check for master or main branch
            assert (
                name is not None
            ), "`name` of the template is required to clone a project."
            _model.info.options["name"] = name
            assert isinstance(repo_or_store, str) and isinstance(branch, str)
            project_clone(
                name,
                Path(output_path),
                repo=repo_or_store,
                branch=branch,
                sparse_checkout=sparse_checkout,
            )
            copy_tree(_model.path, output_path)
        else:
            # works with S3 bucket, haven't failed yet
            assert (
                remotes_config is not None
            ), """\
                `remotes_config` is required in order to pull projects into
                 BentoML modelstore. Refers to
                 https://spacy.io/usage/projects#remote
                 for more information. We will accept remotes
                 as shown:
                 {
                    'remotes':
                    {
                        'default':'s3://spacy-bucket',
                    }
                }
                 """
            os.makedirs(output_path, exist_ok=True)
            with Path(output_path, "project.yml").open("w") as inf:
                yaml.dump(remotes_config, inf)
            for remote in remotes_config.get("remotes", {}):
                for url, res_path in project_pull(  # type: ignore
                    Path(output_path),
                    remote=remote,
                    verbose=verbose,
                ):
                    if url is not None:  # pragma: no cover
                        logger.info(f"Pulled {res_path} from {repo_or_store}")
        return _model.tag


class _SpacyRunner(BaseModelRunner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        gpu_device_id: t.Optional[int],
        vocab: t.Union["Vocab", bool],
        disable: t.Sequence[str],
        exclude: t.Sequence[str],
        config: t.Union[t.Dict[str, t.Any], "Config", None],
        as_tuples: bool,
        batch_size: t.Optional[int],
        component_cfg: t.Optional[t.Dict[str, t.Dict[str, t.Any]]],
        backend_options: t.Optional[str] = "pytorch",
        name: t.Optional[str] = None,
    ):
        super().__init__(tag=tag, name=name)

        self._vocab: t.Union["Vocab", bool] = vocab
        self._disable = disable
        self._exclude = exclude
        self._config = config

        self._backend_options = backend_options
        self._gpu_device_id = gpu_device_id

        self._as_tuples = as_tuples
        self._batch_size = batch_size
        self._component_cfg = component_cfg

    def _configure(self, backend_options: t.Optional[str]) -> None:
        import thinc.util as thinc_util
        import thinc.backends as thinc_backends

        if self._gpu_device_id is not None and thinc_util.prefer_gpu(
            self._gpu_device_id
        ):  # pragma: no cover
            assert backend_options is not None
            if backend_options == "pytorch":
                thinc_backends.use_pytorch_for_gpu_memory()
            else:
                thinc_backends.use_tensorflow_for_gpu_memory()
            thinc_util.require_gpu(self._gpu_device_id)
            thinc_backends.set_gpu_allocator(backend_options)
            thinc_util.set_active_gpu(self._gpu_device_id)
        else:
            thinc_util.require_cpu()

    def _get_pytorch_gpu_count(self) -> t.Optional[int]:
        assert self._backend_options == "pytorch"
        devs = getattr(torch, "cuda").device_count()
        if devs == 0:
            logger.warning("Installation of Torch is not CUDA-enabled.")
            logger.warning(
                _TORCH_TF_WARNING.format(
                    framework="PyTorch",
                    package="torch",
                    link="https://pytorch.org/get-started/locally/",
                )
            )
            return None
        return devs

    def _get_tensorflow_gpu_count(self) -> t.Optional[int]:
        assert self._backend_options == "tensorflow"
        from tensorflow.python.client import (
            device_lib,  # FIXME: using private API; type: ignore
        )

        try:
            return len(
                [
                    x
                    for x in getattr(device_lib, "list_local_devices")()
                    if getattr(x, "device_type") == "GPU"
                ]
            )
        except (AttributeError, Exception):  # pylint: disable=broad-except
            logger.warning(
                _TORCH_TF_WARNING.format(
                    framework="Tensorflow 2.x",
                    package="tensorflow",
                    link="https://www.tensorflow.org/install/gpu",
                )
            )
            return None

    @property
    def num_replica(self) -> int:
        if self.resource_quota.nvidia_gpu:
            if self._backend_options == "pytorch":
                num_devices = self._get_pytorch_gpu_count()
            else:
                num_devices = self._get_tensorflow_gpu_count()
            return num_devices if num_devices is not None else 1
        return 1

    def _setup(self) -> None:
        self._configure(self._backend_options)
        self._model = load(
            self._tag,
            vocab=self._vocab,
            exclude=self._exclude,
            disable=self._disable,
            config=self._config,
            model_store=self.model_store,
        )

    def _run_batch(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        input_data: t.Union[
            t.Sequence[t.Tuple[t.Union[str, "Doc"], t.Any]],
            t.Union[str, "Doc"],
        ],  # type: ignore[reportUnknownParameterType]
    ) -> t.Union[t.Iterator["Doc"], t.Iterator[t.Tuple["Doc", t.Any]]]:  # type: ignore[reportUnknownParameterType]
        return self._model.pipe(  # type: ignore[reportGeneralTypeIssues]
            input_data,  # type: ignore[reportGeneralTypeIssues]
            as_tuples=self._as_tuples,  # type: ignore
            batch_size=self._batch_size,
            disable=self._disable,
            component_cfg=self._component_cfg,
            n_process=self.num_replica,
        )


@inject
def load_runner(
    tag: t.Union[str, Tag],
    *,
    gpu_device_id: t.Optional[int] = None,
    backend_options: t.Optional[str] = None,
    name: t.Optional[str] = None,
    vocab: t.Union["Vocab", bool] = True,  # type: ignore[reportUnknownParameterType]
    disable: t.Sequence[str] = tuple(),
    exclude: t.Sequence[str] = tuple(),
    config: t.Union[t.Dict[str, t.Any], "Config", None] = None,
    as_tuples: bool = False,
    batch_size: t.Optional[int] = None,
    component_cfg: t.Optional[t.Dict[str, t.Dict[str, t.Any]]] = None,
) -> "_SpacyRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. :func:`bentoml.spacy.load_runner` implements a Runner class that
    wrap around :obj:`spacy.language.Language` model, which optimize it for the BentoML runtime.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore..
        gpu_device_id (`int`, `optional`, defaults to `None`):
            GPU device ID.
        backend_options (`str`, `optional`, defaults to `None`):
            Backend options for Thinc. Either "pytorch" or "tensorflow".
        vocab (:code:`Union[spacy.vocab.Vocab, bool]`, `optional`, defaults to `True`):
            Optional vocab to pass in on initialization. If True, a new Vocab object will be created.
        disable (`Sequence[str]`, `optional`):
            Names of pipeline components to disable.
        exclude (`Sequence[str]`, `optional`):
            Names of pipeline components to exclude. Excluded
            components won't be loaded.
        config (:code:`Union[Dict[str, Any], spacy.Config]`, `optional`):
            Config overrides as nested dict or dict
            keyed by section values in dot notation.
        as_tuples (`bool`, `optional`, defaults to `False`):
            If set to True, inputs should be a sequence of
            (text, context) tuples. Output will then be a sequence of
            (doc, context) tuples.
        batch_size (`int`, `optional`, defaults to `None`):
            The number of texts to buffer.
        component_cfg (:code:`Dict[str, :code:`Dict[str, Any]]`, `optional`, defaults to `None`):
            An optional dictionary with extra keyword
            arguments for specific components.

    Returns:
        :obj:`~bentoml._internal.runner.Runner`: Runner instances for the target :mod:`bentoml.sklearn` model

    Examples:

    .. code-block:: python

        import bentoml

        runner = bentoml.sklearn.load_runner("my_model:latest")
        runner.run([[1,2,3,4]])
    """
    return _SpacyRunner(
        tag=tag,
        gpu_device_id=gpu_device_id,
        backend_options=backend_options,
        name=name,
        vocab=vocab,
        disable=disable,
        exclude=exclude,
        config=config,
        as_tuples=as_tuples,
        batch_size=batch_size,
        component_cfg=component_cfg,
    )
